import datetime
import logging
import os
import sys

from itertools import product

import ecmwfapi
import pandas as pd
import xarray as xr

from download_toolbox.cli import CDSDownloadArgParser, DownloadArgParser
from download_toolbox.dataset import DatasetConfig
from download_toolbox.download import ThreadedDownloader, DownloaderError
from download_toolbox.location import Location
from download_toolbox.time import Frequency
"""

"""

class ECMWFDatasetConfig(DatasetConfig):
    CDI_MAP = {
        'tas': '2m_temperature',
        'ta': 'temperature',  # 500
        'tos': 'sea_surface_temperature',
        'ps': 'surface_pressure',
        'zg': 'geopotential',  # 250 and 500
        'hus': 'specific_humidity',  # 1000
        'rlds': 'surface_thermal_radiation_downwards',
        'rsds': 'surface_solar_radiation_downwards',
        'uas': '10m_u_component_of_wind',
        'vas': '10m_v_component_of_wind',
        'ua': 'u_component_of_wind',
        'va': 'v_component_of_wind',
        'sic': 'sea_ice_cover',
        'psl': 'mean_sea_level_pressure',
    }

    def __init__(self,
                 identifier: str = None,
                 cdi_map: object = None,
                 **kwargs):
        super().__init__(identifier="cds"
                         if identifier is None else identifier,
                         **kwargs)

        self._cdi_map = CDSDatasetConfig.CDI_MAP
        if cdi_map is not None:
            self._cdi_map.update(cdi_map)

        for var_config in self.variables:
            if var_config.prefix not in self._cdi_map:
                raise RuntimeError("{} requested but we don't have a map to CDS API naming, "
                                   "please select one of: {}".format(var_config.prefix, self._cdi_map))

    @property
    def cdi_map(self):
        return self._cdi_map


class SEAS5DatasetConfig(ECMWFDatasetConfig):
    """
    ECMWF seasonal forecasts (SEAS5)
    https://www.ecmwf.int/en/newsletter/154/meteorology/ecmwfs-new-long-range-forecasting-system-seas5
    """
    def __init__(self,
                 identifier: str = None,
                 cdi_map: object = None,
                 **kwargs):
        super().__init__(identifier="seas5"
                         if identifier is None else identifier,
                         cdi_map=cdi_map,
                         **kwargs)

class HRESDatasetConfig(ECMWFDatasetConfig):
    """
    ECMWF Atmospheric Model high resolution 15-day forecast (Set I - HRES)
    https://www.ecmwf.int/en/forecasts/datasets/set-i
    """
    def __init__(self,
                 identifier: str = None,
                 cdi_map: object = None,
                 **kwargs):
        super().__init__(identifier="seas5"
                         if identifier is None else identifier,
                         cdi_map=cdi_map,
                         **kwargs)


class HRESDownloader(ClimateDownloader):
    """Climate downloader to provide CMIP6 reanalysis data from ESGF APIs

    :param identifier: how to identify this dataset

    """

    PARAM_TABLE = 128

    # Background on the use of forecast and observational data
    # https://confluence.ecmwf.int/pages/viewpage.action?pageId=85402030
    # https://confluence.ecmwf.int/display/CKB/ERA5%3A+data+documentation#ERA5:datadocumentation-Dateandtimespecification
    HRES_PARAMS = {
        "siconca": (31, "siconc"),  # sea_ice_area_fraction
        "tos": (34, "sst"),  # sea surface temperature (actually
        # sst?)
        "zg": (129, "z"),  # geopotential
        "ta": (130, "t"),  # air_temperature (t)
        "hus": (133, "q"),  # specific_humidity
        "psl": (134, "sp"),  # surface_pressure
        "uas": (165, "u10"),  # 10m_u_component_of_wind
        "vas": (166, "v10"),  # 10m_v_component_of_wind
        "tas": (167, "t2m"),  # 2m_temperature (t2m)
        # https://confluence.ecmwf.int/display/CKB/ERA5%3A+data+documentation#ERA5:datadocumentation-Meanrates/fluxesandaccumulations
        # https://apps.ecmwf.int/codes/grib/param-db/?id=175
        # https://confluence.ecmwf.int/pages/viewpage.action?pageId=197702790
        #
        # Mean rate/flux parameters in ERA5 (e.g. Table 4 for surface and
        # single levels) provide similar information to accumulations (e.g.
        # Table 3 for surface and single levels), except they are expressed as
        # temporal means, over the same processing periods, and so have units
        # of "per second".
        "rlds": (175, "strd"),
        "rsds": (169, "ssrd"),

        # plev  129.128 / 130.128 / 133.128
        # sfc   31.128 / 34.128 / 134.128 /
        #       165.128 / 166.128 / 167.128 / 169.128 / 177.128

        # ORAS5 variables in param-db (need to consider depth)
        # "thetao":       (151129, "thetao"),
        # "so":           (151130, "so"),
        # Better matches than equivalent X / Y parameters in param-db
        # "uo":           (151131, "uo"),
        # "vo":           (151132, "vo"),
    }

    # https://confluence.ecmwf.int/display/UDOC/Keywords+in+MARS+and+Dissemination+requests
    MARS_TEMPLATE = """
retrieve,
  class=od,
  date={date},
  expver=1,
  levtype={levtype},
  {levlist}param={params},
  step={step},
  stream=oper,
  time=12:00:00,
  type=fc,
  area={area},
  grid=0.25/0.25,
  target="{target}",
  format=netcdf
    """

    def __init__(self, *args, identifier: str = "mars.hres", **kwargs):
        super().__init__(*args, identifier=identifier, **kwargs)

        self._server = ecmwfapi.ECMWFService("mars")

    def _single_download(self, var_names: object, pressures: object,
                         req_dates: object):
        """

        :param var_names:
        :param pressures:
        :param req_dates:
        :return:
        """

        for dt in req_dates:
            assert dt.year == req_dates[0].year

        downloads = []
        levtype = "plev" if pressures else "sfc"

        for req_batch in batch_requested_dates(req_dates, attribute="month"):
            req_batch = sorted(req_batch)
            request_month = req_batch[0].strftime("%Y%m")

            logging.info("Downloading month file {}".format(request_month))

            if req_batch[-1] - datetime.datetime.utcnow().date() == \
                    datetime.timedelta(days=-1):
                logging.warning("Not allowing partial requests at present, "
                                "removing {}".format(req_batch[-1]))
                req_batch = req_batch[:-1]

            request_target = os.path.join(
                self.base_path, self.hemisphere_str[0],
                "{}.{}.nc".format(levtype, request_month))

            os.makedirs(os.path.dirname(request_target), exist_ok=True)

            request = self.mars_template.format(
                area="/".join([str(s) for s in self.hemisphere_loc]),
                date="/".join([el.strftime("%Y%m%d") for el in req_batch]),
                levtype=levtype,
                levlist="levelist={},\n  ".format(pressures)
                if pressures else "",
                params="/".join([
                    "{}.{}".format(self.params[v][0], self.param_table)
                    for v in var_names
                ]),
                target=request_target,
                # We are only allowed date prior to -24 hours ago, dynamically
                # retrieve if date is today
                # TODO: too big - step="/".join([str(i) for i in range(24)]),
                step=0,
            )

            if not os.path.exists(request_target):
                logging.debug("MARS REQUEST: \n{}\n".format(request))

                try:
                    self._server.execute(request, request_target)
                except ecmwfapi.api.APIException:
                    logging.exception("Could not complete ECMWF request: {}")
                else:
                    downloads.append(request_target)
            else:
                logging.debug("Already have {}".format(request_target))
                downloads.append(request_target)

        logging.debug("Files downloaded: {}".format(downloads))

        ds = xr.open_mfdataset(downloads)
        ds = ds.resample(time='1D', keep_attrs=True).mean(keep_attrs=True)

        for var_name, pressure in product(
                var_names,
                pressures.split('/') if pressures else [None]):
            var = var_name if not pressure else \
                "{}{}".format(var_name, pressure)

            da = getattr(ds, self.params[var_name][1])

            if pressure:
                da = da.sel(level=int(pressure))

            self.save_temporal_files(var, da)

        ds.close()

        if self.delete:
            for downloaded_file in downloads:
                if os.path.exists(downloaded_file):
                    logging.info("Removing {}".format(downloaded_file))
                    os.unlink(downloaded_file)

    def download(self):
        """

        """
        logging.info("Building request(s), downloading and daily averaging "
                     "from {} API".format(self.identifier.upper()))

        sfc_vars = [
            var for idx, var in enumerate(self.var_names)
            if not self.levels[idx]
        ]
        level_vars = [
            var for idx, var in enumerate(self.var_names) if self.levels[idx]
        ]
        levels = "/".join([
            str(s)
            for s in sorted(set([p for ps in self.levels if ps for p in ps]))
        ])

        # req_dates = self.filter_dates_on_data()

        dates_per_request = \
            batch_requested_dates(self._dates,
                                  attribute=self.group_dates_by)

        for req_batch in dates_per_request:
            if len(sfc_vars) > 0:
                self._single_download(sfc_vars, None, req_batch)

            if len(level_vars) > 0:
                self._single_download(level_vars, levels, req_batch)

        logging.info("{} daily files downloaded".format(
            len(self._files_downloaded)))

    @property
    def mars_template(self):
        return getattr(self, "MARS_TEMPLATE")

    @property
    def params(self):
        return getattr(self, "HRES_PARAMS")

    @property
    def param_table(self):
        return getattr(self, "PARAM_TABLE")

class SEASDownloader(HRESDownloader):
    MARS_TEMPLATE = """
retrieve,
    class=od,
    date={date},
    expver=1,
    levtype={levtype},
    method=1,
    number=0/1/2/3/4/5/6/7/8/9/10/11/12/13/14/15/16/17/18/19/20/21/22/23/24,
    origin=ecmf,
    {levlist}param={params},
    step=0/to/2232/by/24,
    stream=mmsf,
    system=5,
    time=00:00:00,
    type=fc,
    target="{target}",
    format=netcdf,
    grid=0.25/0.25,
    area={area}"""

    def _single_download(self, var_names: object, pressures: object,
                         req_dates: object):
        """

        :param var_names:
        :param pressures:
        :param req_dates:
        :return:
        """

        for dt in req_dates:
            assert dt.year == req_dates[0].year

        downloads = []
        levtype = "plev" if pressures else "sfc"

        for req_date in req_dates:
            request_day = req_date.strftime("%Y%m%d")

            logging.info("Downloading daily file {}".format(request_day))

            request_target = os.path.join(
                self.base_path, self.hemisphere_str[0],
                "{}.{}.nc".format(levtype, request_day))
            os.makedirs(os.path.dirname(request_target), exist_ok=True)

            request = self.mars_template.format(
                area="/".join([str(s) for s in self.hemisphere_loc]),
                date=req_date.strftime("%Y-%m-%d"),
                levtype=levtype,
                levlist="levelist={},\n  ".format(pressures)
                if pressures else "",
                params="/".join([
                    "{}.{}".format(self.params[v][0], self.param_table)
                    for v in var_names
                ]),
                target=request_target,
            )

            if not os.path.exists(request_target):
                logging.debug("MARS REQUEST: \n{}\n".format(request))

                try:
                    self._server.execute(request, request_target)
                except ecmwfapi.api.APIException:
                    logging.exception("Could not complete ECMWF request: {}")
                else:
                    downloads.append(request_target)
            else:
                logging.debug("Already have {}".format(request_target))
                downloads.append(request_target)

        logging.debug("Files downloaded: {}".format(downloads))

        for download_filename in downloads:
            logging.info("Processing {}".format(download_filename))
            ds = xr.open_dataset(download_filename)
            ds = ds.mean("number")

            for var_name, pressure in product(
                    var_names,
                    pressures.split('/') if pressures else [None]):
                var = var_name if not pressure else \
                    "{}{}".format(var_name, pressure)

                da = getattr(ds, self.params[var_name][1])

                if pressure:
                    da = da.sel(level=int(pressure))

                self.save_temporal_files(var, da, date_format="%Y%m%d")

            ds.close()

        if self.delete:
            for downloaded_file in downloads:
                if os.path.exists(downloaded_file):
                    logging.info("Removing {}".format(downloaded_file))
                    os.unlink(downloaded_file)

    def save_temporal_files(self, var, da, date_format=None, freq=None):
        """

        :param var:
        :param da:
        :param date_format:
        :param freq:
        """
        var_folder = self.get_data_var_folder(var)

        req_date = pd.to_datetime(da.time.values[0])
        latlon_path, regridded_name = \
            self.get_req_filenames(var_folder,
                                   req_date,
                                   date_format=date_format)

        logging.info("Retrieving and saving {}".format(latlon_path))
        da.compute()
        da.to_netcdf(latlon_path)

        if not os.path.exists(regridded_name):
            self._files_downloaded.append(latlon_path)


def main(identifier, extra_kwargs=None):
    args = download_args()

    logging.info("ECMWF {} Data Downloading".format(identifier))
    cls = getattr(sys.modules[__name__], "{}Downloader".format(identifier))

    if extra_kwargs is None:
        extra_kwargs = dict()

    instance = cls(
        identifier="mars.{}".format(identifier.lower()),
        var_names=args.vars,
        dates=[
            pd.to_datetime(date).date()
            for date in pd.date_range(args.start_date, args.end_date, freq="D")
        ],
        delete_tempfiles=args.delete,
        levels=args.levels,
        north=args.hemisphere == "north",
        south=args.hemisphere == "south",
        **extra_kwargs)
    instance.download()
    instance.regrid()


def seas_main():
    main("SEAS", dict(group_dates_by="day",))


def hres_main():
    main("HRES")


import datetime as dt
import logging
import os

import pandas as pd

from download_toolbox.dataset import DatasetConfig
from download_toolbox.cli import DownloadArgParser
from download_toolbox.download import ThreadedDownloader, DownloaderError
from download_toolbox.utils import FTPClient
from download_toolbox.location import Location
from download_toolbox.time import Frequency

"""

"""

"""
    https://osi-saf.eumetsat.int/community/list-of-service-messages/release-osi-saf-monthly-mean-sea-ice-concentration-cdricdr
"""
invalid_sic_months = {
    "north": [
        dt.date(1978, 10, 30),
        dt.date(1986, 4, 30),
        dt.date(1986, 5, 31),
        dt.date(1986, 6, 30),
        dt.date(1987, 12, 31),
    ],
    "south": [
        dt.date(1978, 10, 30),
        dt.date(1986, 4, 30),
        dt.date(1986, 5, 31),
        dt.date(1986, 6, 30),
        dt.date(1987, 12, 31),
    ]
}

invalid_sic_days = {
    "north": [
        *[d.date() for d in
          pd.date_range(dt.date(1979, 5, 21), dt.date(1979, 6, 4))],
        *[d.date() for d in
          pd.date_range(dt.date(1979, 6, 10), dt.date(1979, 6, 26))],
        dt.date(1979, 7, 1),
        *[d.date() for d in
          pd.date_range(dt.date(1979, 7, 24), dt.date(1979, 7, 28))],
        *[d.date() for d in
          pd.date_range(dt.date(1980, 1, 4), dt.date(1980, 1, 10))],
        *[d.date() for d in
          pd.date_range(dt.date(1980, 2, 27), dt.date(1980, 3, 4))],
        *[d.date() for d in
          pd.date_range(dt.date(1980, 3, 16), dt.date(1980, 3, 22))],
        *[d.date() for d in
          pd.date_range(dt.date(1980, 4, 9), dt.date(1980, 4, 15))],
        *[d.date() for d in
          pd.date_range(dt.date(1981, 2, 27), dt.date(1981, 3, 5))],
        *[d.date() for d in
          pd.date_range(dt.date(1984, 8, 12), dt.date(1984, 8, 24))],
        dt.date(1984, 9, 14),
        *[d.date() for d in
          pd.date_range(dt.date(1985, 9, 22), dt.date(1985, 9, 28))],
        *[d.date() for d in
          pd.date_range(dt.date(1986, 3, 29), dt.date(1986, 7, 1))],
        *[d.date() for d in
          pd.date_range(dt.date(1987, 1, 3), dt.date(1987, 1, 19))],
        *[d.date() for d in
          pd.date_range(dt.date(1987, 1, 29), dt.date(1987, 2, 2))],
        dt.date(1987, 2, 23),
        *[d.date() for d in
          pd.date_range(dt.date(1987, 2, 26), dt.date(1987, 3, 2))],
        dt.date(1987, 3, 13),
        *[d.date() for d in
          pd.date_range(dt.date(1987, 3, 22), dt.date(1987, 3, 26))],
        *[d.date() for d in
          pd.date_range(dt.date(1987, 4, 3), dt.date(1987, 4, 17))],
        *[d.date() for d in
          pd.date_range(dt.date(1987, 12, 1), dt.date(1988, 1, 12))],
        dt.date(1989, 1, 3),
        *[d.date() for d in
          pd.date_range(dt.date(1990, 12, 21), dt.date(1990, 12, 26))],
        dt.date(1979, 5, 28),
        dt.date(1979, 5, 30),
        dt.date(1979, 6, 1),
        dt.date(1979, 6, 3),
        dt.date(1979, 6, 11),
        dt.date(1979, 6, 13),
        dt.date(1979, 6, 15),
        dt.date(1979, 6, 17),
        dt.date(1979, 6, 19),
        dt.date(1979, 6, 21),
        dt.date(1979, 6, 23),
        dt.date(1979, 6, 25),
        dt.date(1979, 7, 1),
        dt.date(1979, 7, 25),
        dt.date(1979, 7, 27),
        dt.date(1984, 9, 14),
        dt.date(1987, 1, 16),
        dt.date(1987, 1, 18),
        dt.date(1987, 1, 30),
        dt.date(1987, 2, 1),
        dt.date(1987, 2, 23),
        dt.date(1987, 2, 27),
        dt.date(1987, 3, 1),
        dt.date(1987, 3, 13),
        dt.date(1987, 3, 23),
        dt.date(1987, 3, 25),
        dt.date(1987, 4, 4),
        dt.date(1987, 4, 6),
        dt.date(1987, 4, 10),
        dt.date(1987, 4, 12),
        dt.date(1987, 4, 14),
        dt.date(1987, 4, 16),
        dt.date(1987, 4, 4),
        dt.date(1990, 1, 26),
        dt.date(2022, 11, 9),
    ],
    "south": [
        dt.date(1979, 2, 5),
        dt.date(1979, 2, 25),
        dt.date(1979, 3, 23),
        *[d.date() for d in
          pd.date_range(dt.date(1979, 3, 26), dt.date(1979, 3, 30))],
        dt.date(1979, 4, 12),
        dt.date(1979, 5, 16),
        *[d.date() for d in
          pd.date_range(dt.date(1979, 5, 21), dt.date(1979, 5, 27))],
        *[d.date() for d in
          pd.date_range(dt.date(1979, 7, 10), dt.date(1979, 7, 18))],
        dt.date(1979, 8, 10),
        dt.date(1979, 9, 3),
        *[d.date() for d in
          pd.date_range(dt.date(1980, 1, 4), dt.date(1980, 1, 10))],
        dt.date(1980, 2, 16),
        *[d.date() for d in
          pd.date_range(dt.date(1980, 2, 27), dt.date(1980, 3, 4))],
        *[d.date() for d in
          pd.date_range(dt.date(1980, 3, 14), dt.date(1980, 3, 22))],
        dt.date(1980, 3, 31),
        *[d.date() for d in
          pd.date_range(dt.date(1980, 4, 9), dt.date(1980, 4, 15))],
        dt.date(1980, 4, 22),
        *[d.date() for d in
          pd.date_range(dt.date(1981, 2, 27), dt.date(1981, 3, 5))],
        dt.date(1981, 6, 10),
        *[d.date() for d in
          pd.date_range(dt.date(1981, 8, 3), dt.date(1982, 8, 9))],
        dt.date(1982, 8, 6),
        *[d.date() for d in
          pd.date_range(dt.date(1983, 7, 7), dt.date(1983, 7, 11))],
        dt.date(1983, 7, 22),
        dt.date(1984, 6, 12),
        *[d.date() for d in
          pd.date_range(dt.date(1984, 8, 12), dt.date(1984, 8, 24))],
        *[d.date() for d in
          pd.date_range(dt.date(1984, 9, 13), dt.date(1984, 9, 17))],
        *[d.date() for d in
          pd.date_range(dt.date(1984, 10, 3), dt.date(1984, 10, 9))],
        *[d.date() for d in
          pd.date_range(dt.date(1984, 11, 18), dt.date(1984, 11, 22))],
        dt.date(1985, 7, 23),
        *[d.date() for d in
          pd.date_range(dt.date(1985, 9, 22), dt.date(1985, 9, 28))],
        *[d.date() for d in
          pd.date_range(dt.date(1986, 3, 29), dt.date(1986, 11, 2))],
        *[d.date() for d in
          pd.date_range(dt.date(1987, 1, 3), dt.date(1987, 1, 15))],
        *[d.date() for d in
          pd.date_range(dt.date(1987, 12, 1), dt.date(1988, 1, 12))],
        dt.date(1990, 8, 14),
        dt.date(1990, 8, 15),
        dt.date(1990, 8, 24),
        *[d.date() for d in
          pd.date_range(dt.date(1990, 12, 22), dt.date(1990, 12, 26))],
        dt.date(1979, 2, 5),
        dt.date(1979, 2, 25),
        dt.date(1979, 3, 23),
        dt.date(1979, 3, 27),
        dt.date(1979, 3, 29),
        dt.date(1979, 4, 12),
        dt.date(1979, 5, 16),
        dt.date(1979, 7, 11),
        dt.date(1979, 7, 13),
        dt.date(1979, 7, 15),
        dt.date(1979, 7, 17),
        dt.date(1979, 8, 10),
        dt.date(1979, 9, 3),
        dt.date(1980, 2, 16),
        dt.date(1980, 3, 15),
        dt.date(1980, 3, 31),
        dt.date(1980, 4, 22),
        dt.date(1981, 6, 10),
        dt.date(1982, 8, 6),
        dt.date(1983, 7, 8),
        dt.date(1983, 7, 10),
        dt.date(1983, 7, 22),
        dt.date(1984, 6, 12),
        dt.date(1984, 9, 14),
        dt.date(1984, 9, 16),
        dt.date(1984, 10, 4),
        dt.date(1984, 10, 6),
        dt.date(1984, 10, 8),
        dt.date(1984, 11, 19),
        dt.date(1984, 11, 21),
        dt.date(1985, 7, 23),
        *pd.date_range(dt.date(1986, 7, 2), dt.date(1986, 11, 1)),
        dt.date(1990, 8, 14),
        dt.date(1990, 8, 15),
        dt.date(1990, 8, 24),
        dt.date(2022, 11, 9),
    ]
}

var_remove_list = ['time_bnds', 'raw_ice_conc_values', 'total_standard_error',
                   'smearing_standard_error', 'algorithm_standard_error',
                   'status_flag', 'Lambert_Azimuthal_Grid']


class SICDatasetConfig(DatasetConfig):
    def __init__(self,
                 identifier=None,
                 levels=None,
                 var_names=None,
                 **kwargs):
        super().__init__(identifier="osisaf" if identifier is None else identifier,
                         levels=[None] if levels is None else levels,
                         var_names=["siconca"] if var_names is None else var_names,
                         **kwargs)

        invalid_dates = invalid_sic_days if self.frequency == Frequency.DAY else invalid_sic_months
        self._invalid_dates = invalid_dates["north" if self.location.north else "south"]

    @property
    def invalid_dates(self):
        return self._invalid_dates


class SICDownloader(ThreadedDownloader):
    """Downloads OSISAF SIC data from 1978-present using FTP.

    The data comes from individual files, daily or monthly, for this product:
    https://osi-saf.eumetsat.int/products/osi-450-a

    We use the following for FTP downloads:
        - osisaf.met.no

    """
    def __init__(self,
                 dataset: SICDatasetConfig,
                 *args,
                 start_date: object,
                 **kwargs):
        self._conc_start = dt.date(1978, 10, 25)
        self._reproc_start = dt.date(2021, 1, 1)

        if start_date < self._conc_start:
            raise DownloaderError("OSISAF SIC only exists past {}".format(self._conc_start))

        if not (dataset.location.north or dataset.location.south):
            # TODO: other locations are valid, there is work to do to support their "cutting out"
            raise RuntimeError("Please only use this downloader with whole hemispheres, for the mo")

        self._hemi_str = "nh" if dataset.location.north else "sh"

        self._ftp_client = FTPClient(host="osisaf.met.no")

        super().__init__(dataset,
                         *args,
                         # We can download both daily and monthly data
                         source_min_frequency=Frequency.MONTH,
                         start_date=start_date,
                         **kwargs)

    def _single_download(self,
                         var_config: object,
                         req_dates: object) -> list:

        downloaded_files = []

        for file_date in req_dates:
            monthly_file = self.dataset.frequency < Frequency.DAY

            if not monthly_file:
                res_str = ""
                freq_path_str = "{:04d}/{:02d}"
            else:
                res_str = "monthly"
                freq_path_str = "{:04d}"

            # OSI-450-a1
            ftp_conc = "/reprocessed/ice/conc/v3p1/{}/{}".format(res_str, freq_path_str)
            # OSI-430-a
            ftp_reproc = "/reprocessed/ice/conc-cont-reproc/v3p0/{}/{}".format(res_str, freq_path_str)

            if file_date < self._reproc_start:
                file_base = "ice_conc_{}_ease2-250_cdr-v3p0".format(self._hemi_str)
                source_base = ftp_conc
            else:
                file_base = "ice_conc_{}_ease2-250_icdr-v3p0".format(self._hemi_str)
                source_base = ftp_reproc

            source_base = source_base.format(file_date.year, file_date.month)

            if monthly_file:
                source_base = source_base.format(file_date.year)
                dest_base = freq_path_str.format(file_date.year)
                file_in_question = "{}_{:04d}{:02d}.nc". \
                    format(file_base, file_date.year, file_date.month)
            else:
                source_base = source_base.format(file_date.year, file_date.month)
                dest_base = freq_path_str.format(file_date.year, file_date.month)
                file_in_question = "{}_{:04d}{:02d}{:02d}1200.nc". \
                    format(file_base, file_date.year, file_date.month, file_date.day)

            destination_path = os.path.join(var_config.root_path, dest_base, file_in_question)

            if not os.path.exists(os.path.dirname(destination_path)):
                os.makedirs(os.path.dirname(destination_path), exist_ok=True)

            if not os.path.exists(destination_path):
                try:
                    logging.info("Downloading {}".format(destination_path))
                    self._ftp_client.single_request(source_base,
                                                    file_in_question,
                                                    destination_path)
                    downloaded_files.append(destination_path)
                except DownloaderError as e:
                    logging.warning("Failed to download {}: {}".format(destination_path, e))
                    self.missing_dates.append(file_date)
            else:
                logging.debug("{} already exists".format(destination_path))
                downloaded_files.append(destination_path)

        return downloaded_files


def main():
    args = DownloadArgParser().add_workers().parse_args()

    logging.info("OSISAF-SIC Data Downloading")
    location = Location(
        name=args.hemisphere,
        north=args.hemisphere == "north",
        south=args.hemisphere == "south",
    )

    dataset = SICDatasetConfig(
        location=location,
        frequency=getattr(Frequency, args.frequency),
        output_group_by=getattr(Frequency, args.output_group_by),
        config_path=args.config,
        overwrite=args.overwrite_config,
    )

    for start_date, end_date in zip(args.start_dates, args.end_dates):
        logging.info("Downloading between {} and {}".format(start_date, end_date))

        sic = SICDownloader(
            dataset,
            max_threads=args.workers,
            start_date=start_date,
            end_date=end_date,
        )
        sic.download()
        dataset.save_data_for_config(
            rename_var_list=dict(ice_conc="siconca"),
            source_files=sic.files_downloaded,
            var_filter_list=var_remove_list
        )




import datetime as dt
import logging
import re
import requests
import requests.adapters
import os
import zipfile

import cdsapi as cds
import pandas as pd
import xarray as xr

from pprint import pformat
from typing import Union

from download_toolbox.cli import DownloadArgParser, csv_arg
from download_toolbox.dataset import DatasetConfig
from download_toolbox.data.utils import xr_save_netcdf
from download_toolbox.download import ThreadedDownloader, DownloaderError
from download_toolbox.location import Location
from download_toolbox.time import Frequency


class CDSDownloadArgParser(DownloadArgParser):
    def __init__(self,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)

        self.add_argument("-i", "--identifier",
                          help="Name of the output dataset where it's stored, overriding default",
                          default="cds",
                          type=str)

    def add_cds_specs(self):
        """Arguments for dataset and product_type"""
        self.add_argument("-ds", "--dataset",
                          help="Dataset to download",
                          type=str)
        self.add_argument("-pt", "--product-type",
                          help="Product type for the dataset",
                          type=csv_arg,
                          default=None)
        self.add_argument("--time",
                          help="Comma separated list of times for the dataset ('00:00,01:00'...), or 'all' for all 24 hours",
                          type=csv_arg,
                          default=[])

        # TODO: Pull this to constructor and update other downloaders
        self.add_argument("--compress",
                          help="Provide an integer from 1-9 (low to high) on how much to compress the output netCDF",
                          default=None,
                          type=int)
        return self

    def add_derived_specs(self):
        """Arguments for derived datasets"""
        self.add_argument("--daily-statistic",
                          help="Daily statistic for derived datasets",
                          type=str,
                          default="daily_mean")
        self.add_argument("--time-zone",
                          help="Time zone for derived datasets",
                          type=str,
                          default="utc+00:00")
        self.add_argument("--derived-frequency",
                          help="Frequency for derived datasets",
                          type=str,
                          default="1_hourly")
        return self

    def add_var_specs(self):
        super().add_var_specs()
        # TODO: short_names is experimental for CDS downloads only, but maybe should be moved to DatasetConfig
        self.add_argument("-l", "--long-names",
                          help="If provided, this will override name mappings for the given variable prefixes",
                          default=None,
                          type=csv_arg)
        return self


class CDSDatasetConfig(DatasetConfig):
    CDI_MAP = {
        'tas': '2m_temperature',
        'ta': 'temperature',  # 500
        'tos': 'sea_surface_temperature',
        'ps': 'surface_pressure',
        'zg': 'geopotential',  # 250 and 500
        'hus': 'specific_humidity',  # 1000
        'rlds': 'surface_thermal_radiation_downwards',
        'rsds': 'surface_solar_radiation_downwards',
        'uas': '10m_u_component_of_wind',
        'vas': '10m_v_component_of_wind',
        'ua': 'u_component_of_wind',
        'va': 'v_component_of_wind',
        'sic': 'sea_ice_cover',
        'psl': 'mean_sea_level_pressure',
    }

    def __init__(self,
                 identifier: str = None,
                 cdi_map: object = None,
                 # TODO: short_names is experimental for CDS downloads only, but maybe should be moved to DatasetConfig
                 long_names: list = None,
                 **kwargs):
        super().__init__(identifier="cds"
                         if identifier is None else identifier,
                         **kwargs)

        self._cdi_map = CDSDatasetConfig.CDI_MAP
        if cdi_map is not None:
            self._cdi_map.update(cdi_map)

        if long_names is not None:
            self._cdi_map = dict(zip([var_config.prefix for var_config in self.variables], long_names))

        for var_config in self.variables:
            if var_config.prefix not in self._cdi_map:
                raise RuntimeError("{} requested but we don't have a map to CDS API naming, "
                                   "please select one of: {}".format(var_config.prefix, self._cdi_map))

    @property
    def cdi_map(self):
        return self._cdi_map


class ERA5DatasetConfig(CDSDatasetConfig):
    """
    ERA5DatasetConfig - replaced now by CDSDatasetConfig

    Provided for backwards compatibility only
    """
    pass


class CDSDownloader(ThreadedDownloader):
    def __init__(self,
                 dataset: CDSDatasetConfig,
                 *args,
                 show_progress: bool = False,
                 start_date: object,
                 dataset_name: Union[str, None] = None,
                 product_type: Union[list, None] = None,
                 time: Union[list, None] = None,
                 daily_statistic: str = "daily_mean",
                 time_zone: str = "utc+00:00",
                 derived_frequency: str = "1_hourly",
                 compress: Union[int, None] = None,
                 request_args: [dict, None] = None,
                 zipped: bool = False,
                 **kwargs):
        self._client = cds.Client(progress=show_progress)
        self._dataset_name = dataset_name
        self._product_type = product_type
        self._time = time
        # Variables for derived daily statistics
        self._daily_statistic = daily_statistic
        self._time_zone = time_zone
        self._derived_frequency = derived_frequency
        self._compress = compress
        self._request_args = request_args
        self._zipped = zipped

        super().__init__(dataset,
                         *args,
                         source_min_frequency=Frequency.YEAR,
                         # TODO: validate handling of hourly data, but it is
                         #  possible as a temporal resolution
                         source_max_frequency=Frequency.HOUR,
                         start_date=start_date,
                         **kwargs)

        self.download_method = self._single_api_download

        if self.max_threads > 10:
            logging.info("Upping connection limit for max_threads > 10")
            adapter = requests.adapters.HTTPAdapter(
                pool_connections=self.max_threads,
                pool_maxsize=self.max_threads
            )
            self._client.session.mount("https://", adapter)

    def _single_api_download(self,
                             var_config: object,
                             req_dates: object) -> list:
        """Implements a single download from CDS API

        Args:
            var_config:
            req_dates: The requested dates
        """

        logging.debug("Processing {} dates for {}".format(len(req_dates), var_config))
        monthly_request = self.dataset.frequency < Frequency.DAY

        retrieve_ext = "zip" if self._zipped else "nc"
        temp_download_path = os.path.join(var_config.root_path,
                                          self.dataset.location.name,
                                          "temp.{}".format(os.path.basename(
                                              self.dataset.var_filepath(var_config, req_dates,
                                                                        file_extension=retrieve_ext))))
        download_path = os.path.join(var_config.root_path,
                                     self.dataset.location.name,
                                     os.path.basename(self.dataset.var_filepath(var_config, req_dates)))
        os.makedirs(os.path.dirname(download_path), exist_ok=True)

        if os.path.exists(download_path):
            logging.warning(f"We have a downloaded file available, skipping: {download_path}")
            return [download_path]

        # Default to legacy values if not provided
        if not self._product_type:
            product_type = ["reanalysis",] if not monthly_request else ["monthly_averaged_reanalysis_by_hour_of_day",]
        else:
            product_type = self._product_type

        retrieve_dict = {
            "product_type": product_type,
            "variable": self.dataset.cdi_map[var_config.prefix],
            "year": [int(req_dates[0].year),],
            "month": list(set(["{:02d}".format(rd.month)
                               for rd in sorted(req_dates)])),
        }

        if not self._zipped:
            retrieve_dict.update({
                "format": "netcdf",
                "grid": [0.25, 0.25],
                "area": self.dataset.location.bounds,
                "download_format": "unarchived"
            })

        # Add derived dataset-specific keys
        stats_dataset = False
        if self._dataset_name in [
            "derived-era5-pressure-levels-daily-statistics",
            "derived-era5-single-levels-daily-statistics"
        ]:
            stats_dataset = True
            retrieve_dict.update({
                "daily_statistic": self._daily_statistic,
                "time_zone": self._time_zone,
                "frequency": self._derived_frequency
            })

        level_id = "single-levels"
        if var_config.level:
            level_id = "pressure-levels"
            retrieve_dict["pressure_level"] = [var_config.level]

        # Default to legacy values if not provided
        if not self._product_type:
            dataset = "reanalysis-era5-{}{}".format(level_id, "-monthly-means" if monthly_request else "")
        else:
            # TODO: this is a bit of a hack, but it works for now
            # Updating dataset name if multiple pressure levels are requested
            if var_config.level and "single-levels" in self._dataset_name:
                dataset = self._dataset_name.replace("single-levels", "pressure-levels")
            else:
                dataset = self._dataset_name

        # FIXME: this is quite shaky, not using at present
        #    # _, date_end = get_era5_available_date_range(dataset)

        #    # TODO: This updates to dates available for download, prevents
        #    #       redundant downloads but, requires work to prevent
        #    #       postprocess method from running if no downloaded file.
        #    # req_dates = [date for date in req_dates if date <= date_end]
        # END

        if not monthly_request:
            retrieve_dict["day"] = ["{:02d}".format(d) for d in range(1, 32)]

            # No time key required for daily stats dataset, instead uses `time_zone`
            if not stats_dataset:
                if self._time and isinstance(self._time, list):
                    if self._time[0] == "all":
                        time = ["{:02d}:00".format(h) for h in range(0, 24)]
                    else:
                        time = self._time
                else:
                    time = ["12:00",]
                retrieve_dict["time"] = time

        if self._request_args is not None:
            logging.debug("Updating request with dictionary {}".format(self._request_args))
            retrieve_dict.update(self._request_args)

        if os.path.exists(temp_download_path):
            # TODO: we need a better mechanism for keeping temp files and reprocessing
            if self._zipped:
                logging.warning("{} already exists, all we can do is assume to try and reprocess".format(temp_download_path))
            else:
                raise DownloaderError("{} already exists, this shouldn't be the case".format(temp_download_path))
        else:
            try:
                logging.info("Downloading data for {}...".format(var_config.name))
                logging.debug("Request dataset {} with:\n".format(pformat(retrieve_dict)))
                self._client.retrieve(
                    dataset,
                    retrieve_dict,
                    temp_download_path)
                logging.info("Download completed: {}".format(temp_download_path))
            # cdsapi uses raise Exception in many places, so having a catch-all is appropriate
            except Exception as e:
                logging.exception("{} not downloaded, look at the problem".format(temp_download_path))
                self.missing_dates.extend(req_dates)
                return []

        if self._zipped:
            zf = zipfile.ZipFile(temp_download_path)
            zip_output_path = os.path.join(var_config.root_path,
                                           self.dataset.location.name)
            zipped_data_files = [df_name for df_name in zf.namelist()
                                 if df_name.endswith(".nc")
                                 and not os.path.exists(os.path.join(zip_output_path, df_name))]
            zf.extractall(path=zip_output_path,
                          members=zipped_data_files)

            # For the moment we'll keep the zips
            temp_download_path = [os.path.join(zip_output_path, df_name)
                                  for df_name in zf.namelist()
                                  if df_name.endswith(".nc")]
            ds = xr.open_mfdataset(temp_download_path)
        else:
            ds = xr.open_dataset(temp_download_path)

        # TODO: there is duplicated / messy code here from CDS API alterations, clean it up
        # New CDSAPI file holds more data_vars than just variable.
        # Omit them when figuring out default CDS variable name.
        omit_vars = {"number", "expver", "time", "date", "valid_time", "latitude", "longitude", "time_counter_bnds"}
        data_vars = set(ds.data_vars)
        var_list = list(data_vars.difference(omit_vars))
        if not var_list:
            raise ValueError("No variables found in file")
        elif len(var_list) > 1:
            raise ValueError(f"""Multiple variables found in data file!
                                 There should only be one variable.
                                 {var_list}""")
        src_var_name = var_list[0]
        var_name = var_config.name

        # Rename time and variable names for consistency
        rename_vars = {
                       src_var_name: var_name,
                       }
        if "date" in ds:
            rename_vars.update({"date": "time"})
        elif "valid_time" in ds:
            rename_vars.update({"valid_time": "time"})
        elif "time_counter" in ds:
            rename_vars.update({"time_counter": "time"})

        if all([f in ds for f in ["nav_lat", "nav_lon"]]):
            logging.warning("We have nav_lat and nav_lon which suggests a tripolar grid"
                            " which we cannot convert within the downloader yet")
            northing, westing, southing, easting = self.dataset.location.bounds
            # FIXME: naively grab the output based on the latitude
            ds = ds.where(((ds.nav_lat <= northing) & (ds.nav_lat >= southing)).compute(), drop=True)

        da = getattr(ds.rename(rename_vars), var_name)

        # This data downloader handles different pressure_levels in independent
        # files rather than storing them all in separate dimension of one array/file.
        if "pressure_level" in da.dims:
            da = da.squeeze(dim="pressure_level").drop_vars("pressure_level")

        if "number" in da.coords:
            da = da.drop_vars("number")

        # Updating coord attribute definitions (needs file read in with `decode_cf=False`)
        if "coordinates" in da.attrs:
            omit_attrs = ["number", "expver", "isobaricInhPa"]
            attributes = re.sub(r"valid_time|date", "time", da.attrs["coordinates"]).split()
            attributes = [attr for attr in attributes if attr not in omit_attrs]
            da.attrs["coordinates"] = " ".join(attributes)

        # Bryn Note:
        # expver = 1: ERA5
        # expver = 5: ERA5T
        # The latest 3 months of data is ERA5T and may be subject to changes.
        # Data prior to this is from ERA5.
        # The new CDSAPI returns combined data when `reanalysis` is requested.
        if 'expver' in da.coords:
            logging.warning("expver in coordinates, new cdsapi returns ERA5 and "
                            "ERA5T combined, this needs further work: expver needs "
                            "storing for later overwriting")
            # Ref: https://confluence.ecmwf.int/pages/viewpage.action?pageId=173385064
            # da = da.sel(expver=1).combine_first(da.sel(expver=5))
            da = da.drop_vars("expver")
        logging.info("Saving corrected CDS file to {}".format(download_path))
        xr_save_netcdf(da, download_path, complevel=self._compress)
        da.close()

        if type(temp_download_path) is not list:
            temp_download_path = [temp_download_path,]
        for tdp in temp_download_path:
            if os.path.exists(tdp):
                logging.info("Removing {}".format(tdp))
                os.unlink(tdp)

        return [download_path]

    def _single_download(self,
                         var_config: object,
                         req_dates: object) -> list:
        logging.warning("You're not going to get data by calling this! "
                        "Set download_method to an actual implementation.")


def get_era5_available_date_range(dataset: str = "reanalysis-era5-single-levels"):
    """Returns the time range for which ERA5(T) data is available.
    Args:
        dataset: Dataset for which available time range should be returned.
    Returns:
        date_start: Earliest time data is available from.
        date_end: Latest time data available.
    """
    location = f"https://cds.climate.copernicus.eu/api/catalogue/v1/collections/{dataset}"
    res = requests.get(location)

    temporal_interval = res.json()["extent"]["temporal"]["interval"][0]
    time_start, time_end = temporal_interval

    date_start = pd.Timestamp(pd.to_datetime(time_start).date())
    date_end = pd.Timestamp(pd.to_datetime(time_end).date())
    return date_start, date_end


def cds_main():
    args, request_args = (CDSDownloadArgParser().
                          add_var_specs().
                          add_cds_specs().
                          add_derived_specs().
                          add_workers().
                          add_extra_args([
                            (("-z", "--zipped"), dict(action="store_true", default=False,
                                                      help="Zipped version only available, changes request "
                                                           "setup and post-processing prior to save"))
                          ]).
                          parse_known_args())

    logging.info("CDS Data Downloading")

    location = Location(
        name=args.hemisphere,
        north=args.hemisphere == "north",
        south=args.hemisphere == "south",
    )

    dataset = CDSDatasetConfig(
        identifier=args.identifier,
        levels=args.levels,
        location=location,
        var_names=args.vars,
        frequency=getattr(Frequency, args.frequency),
        output_group_by=getattr(Frequency, args.output_group_by),
        config_path=args.config,
        overwrite=args.overwrite_config,
        long_names=args.long_names
    )

    for start_date, end_date in zip(args.start_dates, args.end_dates):
        logging.info("Downloading between {} and {}".format(start_date, end_date))
        cds = CDSDownloader(
            dataset,
            start_date=start_date,
            end_date=end_date,
            max_threads=args.workers,
            request_frequency=getattr(Frequency, args.output_group_by),
            dataset_name=args.dataset,
            product_type=args.product_type,
            time=args.time,
            daily_statistic=args.daily_statistic,
            time_zone=args.time_zone,
            derived_frequency=args.derived_frequency,
            compress=args.compress,
            request_args=request_args,
            zipped=args.zipped
        )
        cds.download()

        dataset.save_data_for_config(
            source_files=cds.files_downloaded,
            var_filter_list=["lambert_azimuthal_equal_area"],
        )





import boto3
import collections
import logging
import os

import pandas as pd
import xarray as xr

from botocore import UNSIGNED
from botocore.config import Config
from typing import Union


def batch_requested_dates(dates: object,
                          attribute: str = "month") -> object:
    """

    TODO: should be using Pandas DatetimeIndexes / Periods for this, but the
     need to refactor slightly, and this is working for the moment

    TODO: we should be yielding from here surely

    :param dates:
    :param attribute:
    :return:
    """
    dates = collections.deque(sorted(dates))

    logging.debug("Got {} dates to batch".format(len(dates)))
    batched_dates = []
    batch = []

    while len(dates):
        if not len(batch):
            batch.append(dates.popleft())
        else:
            if getattr(batch[-1], attribute) == getattr(dates[0], attribute):
                batch.append(dates.popleft())
            else:
                logging.debug("Appending batch of length {}".format(len(batch)))
                batched_dates.append(batch)
                batch = []

    if len(batch):
        batched_dates.append(batch)

    if len(dates) > 0:
        raise RuntimeError("Batching didn't work!")

    logging.debug("Return {} batches for {} batch".format(len(batched_dates), attribute))
    return batched_dates


def merge_files(new_datafile: object,
                other_datafile: object,
                drop_variables: object = None):
    """

    :param new_datafile:
    :param other_datafile:
    :param drop_variables:
    """
    drop_variables = list() if drop_variables is None else drop_variables

    if other_datafile is not None:
        (datafile_path, new_filename) = os.path.split(new_datafile)
        moved_new_datafile = \
            os.path.join(datafile_path, "new.{}".format(new_filename))
        os.rename(new_datafile, moved_new_datafile)
        d1 = xr.open_dataarray(moved_new_datafile,
                               drop_variables=drop_variables)

        logging.debug("Concatenating with previous data {}".format(
            other_datafile
        ))
        d2 = xr.open_dataarray(other_datafile,
                               drop_variables=drop_variables)

        # Ordering is important for merging, place the temporary over the existing
        # as it will ensure that code updates and other changes are updated
        new_ds = xr.concat([d2, d1], dim="time").\
            sortby("time").\
            drop_duplicates("time", keep="first")

        logging.info("Saving merged data to {}... ".
                     format(new_datafile))
        new_ds.to_netcdf(new_datafile)
        os.unlink(other_datafile)
        os.unlink(moved_new_datafile)


def s3_file_download(bucket_name: str, key: str, filename: str) -> None:
    """
    Download a file from S3 bucket to local storage.

    If the file already exists and matches the expected size, skip download.
    If the file exists but is incomplete, re-download it.

    Args:
        bucket_name: Name of the S3 bucket.
        key: Key of the file in the S3 bucket.
        filename: Local path to save the downloaded file.
    """
    s3 = boto3.client("s3", config=Config(signature_version=UNSIGNED))

    # Get file size from S3 bucket
    try:
        metadata = s3.head_object(Bucket=bucket_name, Key=key)
        s3_file_size = metadata['ContentLength']
    except Exception as e:
        logging.error(f"Error getting metadata: {e}")
        return

    # Check if local file exists and matches the expected size
    if os.path.exists(filename):
        local_file_size = os.path.getsize(filename)
        if local_file_size == s3_file_size:
            logging.info(f"File already downloaded: {filename}")
            return
        else:
            logging.info(f"Incomplete file found (local: {local_file_size}, remote: {s3_file_size}). Re-downloading...")

    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, "wb") as f:
        s3.download_fileobj(bucket_name, key, f)
    logging.info(f"Download complete: {filename}")


def xr_save_netcdf(da: xr.DataArray, file_path: str, complevel: int = 0) -> None:
    """
    Save xarray Dataarray to netCDF file with optional compression.

    Args:
        da: The xarray dataarray to be output to netCDF.
        file_path: Path to save the netCDF file.
        complevel (optional): Level of compression to apply.
                              Defaults to 0.
    """
    if complevel:
        compression = dict(zlib=True, complevel=int(complevel))
        var_encoding = {da.name: compression}
        coords_encoding = {coord: compression for coord in da.coords}
        da.to_netcdf(file_path, mode="w", encoding=var_encoding | coords_encoding)
    else:
        var_encoding = {da.name: {}}
        coords_encoding = {coord: {} for coord in da.coords}
        da.to_netcdf(file_path, mode="w", encoding=var_encoding | coords_encoding)



import datetime as dt
import ftplib
import importlib
import logging
import subprocess as sp
import sys
import threading
from ftplib import FTP

from functools import wraps

import dask
import requests
from dask.distributed import Client, LocalCluster


def get_implementation(location):
    if ":" not in location:
        if hasattr(sys.modules[__name__], location):
            return getattr(sys.modules[__name__], location)
        else:
            raise ImportError("There is no {} available in sys.modules[__name__] "
                              "and no module path provided".format(location))
    module_ref, object_name = location.split(":")
    implementation = None

    try:
        module = importlib.import_module(module_ref)
        implementation = getattr(module, object_name)
    except ImportError:
        logging.exception("Unable to import from location: {}".format(location))

    return implementation


def run_command(command: str, dry: bool = False):
    """Run a shell command

    A wrapper in case we want some additional handling to go in here

    :param command:
    :param dry:
    :return:

    """
    if dry:
        logging.info("Skipping dry commaand: {}".format(command))
        return 0

    ret = sp.run(command, shell=True)
    if ret.returncode < 0:
        logging.warning("Child was terminated by signal: {}".
                        format(-ret.returncode))
    else:
        logging.info("Child returned: {}".format(-ret.returncode))

    return ret


# This is adapted from the data/loaders implementations
class DaskWrapper:
    """

    :param dask_port:
    :param dask_timeouts:
    :param dask_tmp_dir:
    :param workers:
    """

    def __init__(self,
                 dask_port: int = 8888,
                 dask_timeouts: int = 60,
                 dask_tmp_dir: object = "/tmp",
                 workers: int = 8,
                 scheduler: str = "single-threaded",
                 ):

        self._dashboard_port = dask_port
        self._timeout = dask_timeouts
        self._tmp_dir = dask_tmp_dir
        self._workers = workers
        self._scheduler = scheduler

        self._cluster = None
        self._client = None

    def __enter__(self):
        """

        """
        dashboard = "localhost:{}".format(self._dashboard_port)

        with dask.config.set({
            "temporary_directory": self._tmp_dir,
            "distributed.comm.timeouts.connect": self._timeout,
            "distributed.comm.timeouts.tcp": self._timeout,
            # "scheduler": self._scheduler, # Fix to "single-threaded" for netCDF4 >=1.6.1 not thread-safe.
        }):
            self._cluster = LocalCluster(
                dashboard_address=dashboard,
                n_workers=self._workers,
                threads_per_worker=1,
                scheduler_port=0,
            )
            logging.info("Dashboard at {}".format(dashboard))

            self._client = Client(self._cluster)
            logging.info("Using dask client {}".format(self._client))

        return self

    def __exit__(self, *exc_details):
        self._client.close()
        self._client = None
        # TODO: Leaving the cluster alive?

class FTPClient(object):
    def __init__(self,
                 host: str,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)

        self._ftp = None
        self._ftp_host = host
        self._cache = dict()
        self._ftp_connections = dict()

    def single_request(self,
                       source_dir: object,
                       source_filename: list,
                       destination_path: object):
        thread_id = threading.get_native_id()
        if threading.get_native_id() not in self._ftp_connections:
            logging.debug("FTP opening for thread {}".format(thread_id))
            self._ftp_connections[thread_id] = FTP(self._ftp_host)
            ftp_connection = self._ftp_connections[thread_id]
            ftp_connection.login()
        else:
            ftp_connection = self._ftp_connections[thread_id]

        try:
            logging.debug("FTP changing to {}".format(source_dir))
            # self._ftp.cwd(source_dir)

            if source_dir not in self._cache:
                self._cache[source_dir] = ftp_connection.nlst(source_dir)

            ftp_files = [el for el in self._cache[source_dir] if el.endswith(source_filename)]
            if not len(ftp_files):
                raise ClientError("File is not available: {}".format(source_filename))
        except ftplib.error_perm as e:
            raise ClientError("FTP error, possibly missing directory {}: {}".format(source_dir, e))

        logging.debug("FTP Attempting to retrieve to {} from {}".format(destination_path, ftp_files[0]))
        with open(destination_path, "wb") as fh:
            ftp_connection.retrbinary("RETR {}".format(ftp_files[0]), fh.write)


class HTTPClient(object):
    def __init__(self,
                 host: str,
                 *args,
                 source_base: object = None,
                 **kwargs):
        super().__init__(*args, **kwargs)

        self._host = host
        self._source_base = source_base

    def single_request(self,
                       source: object,
                       destination_path: object,
                       method: str = "get",
                       request_options: dict = None):
        request_options = dict() if request_options is None else request_options
        source_url = "/".join([self._host, self._source_base, source])

        try:
            logging.debug("{}-ing {} with {}".format(method, source_url, request_options))
            response = getattr(requests, method)(source_url, **request_options)
        except requests.exceptions.RequestException as e:
            raise ClientError("HTTP error {}: {}".format(source_url, e))

        if hasattr(response, "status_code") and response.status_code == 200:
            logging.debug("Attempting to output response content to {}".format(destination_path))
            with open(destination_path, "wb") as fh:
                fh.write(response.content)
        else:
            raise ClientError("HTTP response from {} not successful, writing nothing: {}".format(source_url, response.status_code))


class ClientError(RuntimeError):
    pass



import argparse
import concurrent
import datetime as dt
import logging
import os
import re
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor
from functools import lru_cache

import boto3
import xarray as xr
from botocore import UNSIGNED
from botocore.config import Config
from functools import partial

from download_toolbox.cli import DownloadArgParser
from download_toolbox.data.cds import CDSDatasetConfig
from download_toolbox.data.utils import batch_requested_dates, s3_file_download, xr_save_netcdf
from download_toolbox.dataset import DatasetConfig
from download_toolbox.download import DownloaderError, ThreadedDownloader
from download_toolbox.location import Location
from download_toolbox.time import Frequency


class AWSDownloadArgParser(DownloadArgParser):
    """Arguments for AWS datasets"""
    def __init__(self,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)

    def add_aws_specs(self):
        self.add_argument("--delete-cache",
                          help="Delete raw source download cached files after saving output netCDFs",
                          default=False,
                          action=argparse.BooleanOptionalAction)
        self.add_argument("--cache-only",
                          help="Only download the source files into the filecache, do nothing else",
                          default=False,
                          action=argparse.BooleanOptionalAction)

        # TODO: Pull this to constructor and update other downloaders
        self.add_argument("--compress",
                          help="Provide an integer from 1-9 (low to high) on how much to compress the output netCDF",
                          default=None,
                          type=int)

        return self


class AWSDatasetConfig(DatasetConfig):
    # Ref: https://confluence.ecmwf.int/display/CKB/ERA5%3A+data+documentation
    # Map of CMIP6 variable names to ECMWF Grib parameter names
    CMIP6_MAP = {
        "tas":  {"id": 167, "short_name": "2t",           # Near-surface air temperature (CMIP6: tas, ECMWF ID: 167)
                "product_type": "reanalysis", "dataset": "surface-level"},
        "ta":   {"id": 130, "short_name": "t",            # Air temperature at various levels (CMIP6: ta, ECMWF ID: 130)
                "product_type": "reanalysis", "dataset": "pressure-level"},
        "tos":  {"id": 34,  "short_name": "sstk",         # Sea Surface Temperature (CMIP6: tos, ECMWF ID: 34)
                "product_type": "reanalysis", "dataset": "surface-level"},
        "ps":   {"id": 134, "short_name": "sp",           # Surface pressure (CMIP6: ps, ECMWF ID: 134)
                "product_type": "reanalysis", "dataset": "surface-level"},
        # Has to be manually processed from geopotential to geopotential height in post-processing
        "zg":   {"id": 129, "short_name": "z",            # Geopotential height (CMIP6: zg, ECMWF ID: 129)
                "product_type": "reanalysis", "dataset": "pressure-level"},
        "hus":  {"id": 133, "short_name": "q",            # Specific humidity (CMIP6: hus, ECMWF ID: 133)
                "product_type": "reanalysis", "dataset": "pressure-level"},
        # Ref: https://codes.ecmwf.int/grib/param-db/175
        # No reanalysis for this param in AWS ERA5 dataset, only forecast
        # "rlds": {"id": 175, "short_name": "strd"},        # Downward longwave radiation flux at surface (CMIP6: rlds, ECMWF ID: 175)
        #         "product_type": "forecast", "dataset": "accumulation"},
        # Ref: https://codes.ecmwf.int/grib/param-db/169
        # No reanalysis for this param in AWS ERA5 dataset, only forecast
        # "rsds": {"id": 169, "short_name": "ssrd",         # Downward shortwave radiation flux at surface (CMIP6: rsds, ECMWF ID: 169)
        #         "product_type": "forecast", "dataset": "accumulation"},
        "uas":  {"id": 165, "short_name": "10u",          # 10m U-component of wind (CMIP6: uas, ECMWF ID: 165)
            "product_type": "reanalysis", "dataset": "surface-level"},
        "vas":  {"id": 166, "short_name": "10v",          # 10m V-component of wind (CMIP6: vas, ECMWF ID: 166)
            "product_type": "reanalysis", "dataset": "surface-level"},
        "ua":   {"id": 131, "short_name": "u",            # U-component of wind at specific levels (CMIP6: ua, ECMWF ID: 131)
            "product_type": "reanalysis", "dataset": "pressure-level"},
        "va":   {"id": 132, "short_name": "v",            # V-component of wind at specific levels (CMIP6: va, ECMWF ID: 132)
            "product_type": "reanalysis", "dataset": "pressure-level"},
        "sic":  {"id": 31,  "short_name": "ci",           # Sea ice concentration (CMIP6: sic, ECMWF ID: 262001)
            "product_type": "reanalysis", "dataset": "surface-level"},
        "psl":  {"id": 151, "short_name": "msl",          # Sea level pressure (CMIP6: psl, ECMWF ID: 151)
            "product_type": "reanalysis", "dataset": "surface-level"},
    }

    def __init__(self,
                 identifier: str = None,
                 cmip6_map: object = None,
                 **kwargs):
        super().__init__(identifier="aws"
                         if identifier is None else identifier,
                         **kwargs)

        self.cmip6_map = AWSDatasetConfig.CMIP6_MAP
        if cmip6_map is not None:
            self._cmip6_map.update(cmip6_map)

        for var_config in self.variables:
            if var_config.prefix not in self._cmip6_map:
                raise RuntimeError("{} requested but we don't have a map to CDS API naming, "
                                   "please select one of: {}".format(var_config.prefix, self._cmip6_map))

    @property
    def cmip6_map(self):
        return self._cmip6_map

    @cmip6_map.setter
    def cmip6_map(self, value):
        self._cmip6_map = value


class AWSDownloader(ThreadedDownloader):
    """
    Downloader class for retrieving ERA5 climate reanalysis data from AWS.

    Inherits from `ThreadedDownloader`, allowing concurrent downloads of variables
    and date configurations.
    """
    def __init__(self,
                 dataset: CDSDatasetConfig,
                 *args,
                 start_date: object,
                 end_date: object,
                 delete_cache: bool = False,
                 cache_only: bool = False,
                 compress: int = 0,
                 **kwargs):
        """
        Initialise the AWSDownloader instance.

        Args:
            dataset: Dataset configuration object.
            *args: Additional positional arguments to pass to the base class.
            start_date: Start date for the data to download.
            end_date: End date for the data to download.
            delete_cache: If `True`, delete the cache after download.
            cache_only: If `True`, only download files to cache and return.
            compress: Compression level for saved NetCDF files.
                      0 or `None` for no compression.
                      (Defaults to 0. ).
            **kwargs: Additional keyword arguments passed to the base class.

        Raises:
            DownloaderError: If `start_date` or `end_date` are outside AWS ERA5 data range.
        """
        # Date ranges available from AWS data
        era5_start = dt.date(1940, 1, 1)
        era5_end = dt.date(2024, 12, 31)
        logging.getLogger("cdsapi").setLevel(logging.WARNING)

        if start_date < era5_start:
            raise DownloaderError("{} is before the limited ERA5 date available from AWS of {}".
                                  format(start_date, era5_start))
        elif end_date > era5_end:
            raise DownloaderError("{} is after the limited ERA5 date available from AWS of {}".
                                  format(end_date, era5_end))

        super().__init__(dataset,
                         *args,
                         source_min_frequency=Frequency.YEAR,
                         # TODO: validate handling of hourly data, but it is
                         #  possible as a temporal resolution
                         source_max_frequency=Frequency.HOUR,
                         start_date=start_date,
                         end_date=end_date,
                         **kwargs)

        self.download_method = self._single_api_download
        self.product_type_map = self.__product_type_map()
        self.dataset_map = self.__dataset_map()
        self.delete_cache = delete_cache
        self.cache_only = cache_only
        self.compress = compress

    @staticmethod
    def __product_type_map() -> dict:
        """
        Get mapping of short codes for product types.

        Returns:
            Product type mapping with keys like "reanalysis", "forecast", etc.

        Notes:
            * Reference following ECMWF Docs on documentation details
            * https://confluence.ecmwf.int/pages/viewpage.action?pageId=85402030#ERA5terminology:analysisandforecast;timeandsteps;instantaneousandaccumulatedandmeanratesandmin/maxparameters-Analysisandforecast
            * Including difference between 'an' and 'fc'
        """
        return {
            "reanalysis": {
                "short-code": "an",
                "help": (
                    "ERA5 Reanalysis. An analysis of the atmospheric conditions is a blend "
                    "of observations with a previous forecast."
                )
            },
            "forecast": {
                "short-code": "fc.sfc",
                "help": (
                    "ERA5 Forecast Data. A forecast starts with an analysis at a specific time "
                    "(the 'initialisation time'), and a model computes the atmospheric conditions "
                    "for a number of 'forecast steps', at increasing 'validity times', into the future."
                )
            },
            "invariant": {
                "short-code": "invariant",
                "help": (
                    "Variables that don't change over time (e.g. land-sea mask, topography, surface type)."
                )
            }
        }

    @staticmethod
    def __dataset_map() -> dict:
        """
        Get mapping of dataset type long name to ECMWF short-code and product types.

        Returns:
            Dataset type mapping for ERA5 datasets.
        """
        return {
            "pressure-level": {
                "short-code": "pl",
                "product-type": "reanalysis",
                "help": (
                    "Pressure Level data. Variables available on standard pressure levels in the atmosphere "
                    "(e.g. 850 hPa, 500 hPa), such as temperature, geopotential, wind, etc."
                )
            },
            "surface-level": {
                "short-code": "sfc",
                "product-type": "reanalysis",
                "help": (
                    "Surface Level data. Variables at the surface or near-surface, like 2m temperature, "
                    "10m wind, surface pressure, etc."
                )
            },
            "vertically-integrated": {
                "short-code": "vinteg",
                "product-type": "reanalysis",
                "help": (
                    "Vertically Integrated variables. These are quantities integrated through the depth of "
                    "the atmosphere, such as total column water vapor or total column ozone."
                )
            },
            "accumulation": {
                "short-code": "accumu",
                "product-type": "forecast",
                "help": (
                    "Accumulated Forecast Fields. Variables that accumulate over a time interval, "
                    "such as precipitation, snowfall, or runoff."
                )
            },
            "instantaneous": {
                "short-code": "instan",
                "product-type": "forecast",
                "help": (
                    "Instantaneous Forecast Fields. Snapshot values at a specific forecast time, "
                    "e.g. 2m temperature or surface pressure."
                )
            },
            "meanflux": {
                "short-code": "meanflux",
                "product-type": "forecast",
                "help": (
                    "Mean Flux Forecast Fields. Time-averaged fluxes such as sensible heat flux, "
                    "latent heat flux, or radiation components."
                )
            },
            "minmax": {
                "short-code": "minmax",
                "product-type": "forecast",
                "help": (
                    "Minimum/Maximum Forecast Fields. Extremes of a variable over a time period, "
                    "e.g. daily maximum temperature or minimum relative humidity."
                )
            },
            "invariant": {
                "short-code": "invariant",
                "product-type": "invariant",
                "help": (
                    "Invariant Fields. Static variables that do not change over time, "
                    "such as land-sea mask, topography, or surface type."
                )
            }
        }

    @staticmethod
    @lru_cache
    def __list_matching_files(prefix: str,
                              start_date: dt.datetime,
                              end_date: dt.datetime,
                              cmip6_variable: str,
                              ecmwf_variable: str,
                              bucket_name: str,
                              multiple_levels: bool,
                              ) -> dict:
        """
        AWS S3 file paths matching date range and variable filters.

        Args:
            prefix: S3 prefix path for files.
            start_date: Start datetime to filter.
            end_date: End datetime to filter.
            cmip6_variable: Variable name in CMIP6 format.
            ecmwf_variable: Variable name in ECMWF format.
            bucket_name: AWS S3 bucket name.
            multiple_levels: Whether variable is multi-level (pressure level) or not.

        Returns:
            Dictionary of matching files grouped by variable.
        """
        s3 = boto3.resource("s3", config=Config(signature_version=UNSIGNED))
        bucket = s3.Bucket(bucket_name)
        matching_files = defaultdict(list)

        current = start_date.replace(day=1)
        while current <= end_date:
            year_month = current.strftime("%Y%m")
            full_prefix = f"{prefix}{year_month}/"
            for obj in bucket.objects.filter(Prefix=full_prefix):
                nc_file_path = obj.key
                # TODO: The filename seems to follow this pattern:
                # {dataset}.{grib_table}_{parameter_id}_{short_name}.{grid_config}.{start_datetime}_{end_datetime}.nc
                if not nc_file_path.endswith(".nc"):
                    continue
                # Filter by date range
                timestamp_part = nc_file_path.split('.')[-2]
                file_start_date, file_end_date = timestamp_part.split('_')
                try:
                    file_start_date = dt.datetime.strptime(file_start_date, "%Y%m%d%H")
                    file_end_date = dt.datetime.strptime(file_end_date, "%Y%m%d%H")
                except ValueError:
                    continue

                if multiple_levels:
                    # Stores variables with multiple levels in separate daily files
                    # e5.oper.an.pl
                    if not (start_date <= file_start_date <= end_date):
                        continue
                else:
                    # Stores surface variables in separate monthly files
                    # e5.oper.an.sfc
                    if not (start_date <= end_date):
                        continue
                # Filter by parameter
                pattern = rf"\.(\d+_\d+_{ecmwf_variable})\." # Get the parameter details section of filename
                match = re.search(pattern, nc_file_path)
                if not match:
                    continue
                grib_table, parameter_id, ecmwf_short_name = match.group(1).split("_")
                matching_files[cmip6_variable].append(nc_file_path)
            # Move to the next month (even if no. of days less than 31 days)
            current += dt.timedelta(days=32)
            current = current.replace(day=1)

        return matching_files

    def download(self):
        """
        Perform multi-threaded download of ERA5 data from AWS.

        Collects variable and date configurations, batches them for concurrent downloading,
        and uses `ProcessPoolExecutor` to perform downloads in parallel.

        Returns nothing, relies on _single_download to implement
        appropriate updates to this object to record state changes arising from
        downloading.

        Updates the internal `_files_downloaded` list with paths to downloaded files.
        """

        logging.info("Building request(s), downloading and averaging "
                     "from {} API".format(self.dataset.identifier.upper()))

        req_list = list()
        var_config_collection = defaultdict(list)

        # Collate variables with different pressure levels together
        # Avoids needing to download the same file repeatedly from AWS.
        for var_config in self.dataset.variables:
            dates = self.dataset.filter_extant_data(var_config, self.dates)

            for req_date_batch in batch_requested_dates(dates=dates, attribute=self.request_frequency.attribute):
                logging.info("Processing single download for {} with {} dates".
                             format(var_config.name, len(req_date_batch)))

                var_config_collection[var_config.prefix].append((var_config, req_date_batch))

        for var_collection in var_config_collection.values():
            req_list.append([var_collection])

        max_workers = min(len(req_list), self._max_threads)

        if max_workers > 0:
            logging.info("Creating thread pool with {} workers to service {} batches"
                         .format(max_workers, len(req_list)))

            with ProcessPoolExecutor(max_workers=max_workers) as executor:
                futures = []

                for args in req_list:
                    future = executor.submit(self.download_method, *args)
                    futures.append(future)

                for future in concurrent.futures.as_completed(futures):
                    try:
                        files_downloaded = future.result()

                        if files_downloaded is not None:
                            logging.info("{} files downloaded".format(len(files_downloaded)))
                            self._files_downloaded.extend(files_downloaded)
                        else:
                            logging.warning("Nothing downloaded from threaded batch")

                    except Exception as e:
                        logging.exception("Thread failure: {}".format(e))

        logging.info("{} files downloaded".format(len(self._files_downloaded)))

    @staticmethod
    def _preprocess(ds: xr.Dataset,
                    start_dt: dt.datetime,
                    end_dt: dt.datetime,
                    level: int,
                    bounds: list,
                    ) -> xr.Dataset:
        """
        Preprocess individual xarray datasets before combining.

        This method performs the following preprocessing steps:
        - Selects the appropriate pressure level (if available).
        - Filters data based on the given time range.
        - Rolls the longitude coordinate to center the 0° longitude.
        - Extracts a specific geographical region based on the provided latitude and longitude bounds.

        Args:
            ds: Dataset to be processed.
            start_dt: The start datetime for the time range to filter.
            end_dt: The end datetime for the time range to filter.
            level: The pressure level to select if multiple levels are available.
            bounds: A list containing the bounds for the region to extract,
                    in the order [max_lat, min_lon, min_lat, max_lon].

        Returns:
            Processed Dataset with specified region and time range.
        """
        # Extract pressure level
        if "level" in ds.dims:
            # Clearly have multiple pressure levels
            ds = ds.sel(level=level).drop_vars("level")
        else:
            # Surface level data
            ds = ds.sel(time=slice(start_dt, end_dt + dt.timedelta(hours=23)))

        # Extract region
        max_lat, min_lon, min_lat, max_lon = bounds
        lon_mask = (ds.longitude <= max_lon) | (ds.longitude >= min_lon)
        lat_mask = (ds.latitude <= max_lat) & (ds.latitude >= min_lat)
        ds_region = ds.sel(longitude=lon_mask, latitude=lat_mask)

        return ds_region

    def _single_api_download(self,
                            args: list,
                            ) -> list:
        """
        Perform a single-process download batch from AWS based on variable and date ranges.

        Args:
            args: A list of tuples of (var_config, req_dates) for download.

        Returns:
            List of paths to successfully downloaded files.

        Raises:
            DownloaderError: If an unsupported frequency is encountered.
            ValueError: If pressure levels are used with an unsupported dataset.
        """
        # TODO: Add monthly request handling
        #       for AWS data, this is not currently supported
        #       as the data is not available in monthly files
        monthly_request = self.dataset.frequency < Frequency.DAY
        if monthly_request:
            raise DownloaderError("Monthly requests are not supported for AWS data, use `download_cds` instead")

        bucket_name = "nsf-ncar-era5"
        product_type_map = self.__product_type_map()
        dataset_map = self.__dataset_map()

        # Extract root_path from dataset config
        temp_download_path = os.path.join(self.dataset._root_path, "cache")

        # Loop through different pressure levels
        downloaded_paths = []
        downloaded_files = []
        for var_levels in args:
            var_config, req_dates = var_levels
            logging.debug("Processing {} dates for {}".format(len(req_dates), var_config))
            start_date = req_dates[0]
            end_date = req_dates[-1]

            start_dt = dt.datetime.combine(start_date, dt.time(0, 0))
            end_dt = dt.datetime.combine(end_date, dt.time(0, 0))

            # Retrieve filtered file list
            cmip6_variable_code = var_config.prefix
            ecmwf_variable_code = self.dataset.cmip6_map[cmip6_variable_code]["short_name"]

            product = self.dataset.cmip6_map[cmip6_variable_code]["product_type"]
            dataset = self.dataset.cmip6_map[cmip6_variable_code]["dataset"]

            product_code = product_type_map[product]["short-code"]
            dataset_code = dataset_map[dataset]["short-code"]

            logging.info(f"Selected ERA5 product type: {product_code}")
            logging.info(f"Selected ERA5 dataset type: {dataset_code}")

            # Parse prefix
            if product_code == "invariant":
                prefix = f"e5.oper.{product_code}/"
            else:
                prefix = f"e5.oper.{product_code}.{dataset_code}/"

            level = var_config.level
            if level and dataset != "pressure-level":
                raise ValueError(f"Level `{level}` is not supported for `{dataset}` dataset type, "
                                    "this is a surface or near-surface variable"
                )

            filtered_files = self.__list_matching_files(prefix, start_dt, end_dt,
                                cmip6_variable_code, ecmwf_variable_code, bucket_name,
                                multiple_levels=True if level else False)
            joined_files = "\n\t".join(filtered_files[cmip6_variable_code])
            logging.debug(f"Files to download:\n\t{joined_files}")



            download_path = os.path.join(var_config.root_path,
                                        self.dataset.location.name,
                                        os.path.basename(self.dataset.var_filepath(var_config, req_dates)))

            os.makedirs(os.path.dirname(temp_download_path), exist_ok=True)
            os.makedirs(os.path.dirname(download_path), exist_ok=True)

            cached_files = []
            for filtered_file in filtered_files[cmip6_variable_code]:
                cached_file = os.path.join(temp_download_path, os.path.basename(filtered_file))
                s3_file_download(bucket_name, filtered_file, cached_file)
                cached_files.append(cached_file)
                downloaded_files.append(cached_file)

            dataset_preprocess = partial(self._preprocess,
                                               start_dt=start_dt,
                                               end_dt=end_dt,
                                               level=level,
                                               bounds=self.dataset.location.bounds,
                                               )

            try:
                logging.info(f"Downloading data for {var_config.name}...")
                logging.debug(f"Request file:\n{filtered_files[cmip6_variable_code]}")

                ds = xr.open_mfdataset(
                    cached_files,
                    data_vars="minimal",
                    coords="minimal",
                    combine="by_coords",
                    engine="h5netcdf",
                    preprocess=dataset_preprocess,
                    parallel=True,
                    chunks={"time": 24},
                    )

                if self.cache_only:
                    continue
            except Exception as e:
                logging.exception("{} not downloaded, look at the problem".format(temp_download_path))
                self.missing_dates.extend(req_dates)
                continue

            # Roll the data to have the 0 degree longitude at the center
            ds.coords["longitude"] = (ds.coords["longitude"] + 180) % 360 - 180
            ds = ds.sortby(ds.longitude)

            # Figure out the data variable name.
            # It should have the following three dimensions by this point:
            expected_dims = ["time", "latitude", "longitude"]
            for var in ds.data_vars:
                var_dims = ds[var].dims
                if all([dim in var_dims for dim in expected_dims]):
                    src_var_name = var
                    break

            var_name = var_config.name
            rename_vars = {src_var_name: var_name}
            da = getattr(ds.rename(rename_vars), var_name)

            logging.info("Saving corrected ERA5 file to {}".format(download_path))
            xr_save_netcdf(da, download_path, complevel=self.compress)
            ds.close()

            downloaded_paths.append(download_path)

        # Delete cached files if requested
        if self.delete_cache:
            for cached_file in downloaded_files:
                if os.path.exists(cached_file):
                    os.remove(cached_file)

        return downloaded_paths

    def _single_download(self,
                         *args: list) -> list:
        raise RuntimeError("You're not going to get data by calling this! "
                           "Set download_method to an actual implementation.")


def main():
    args = AWSDownloadArgParser().add_var_specs().add_aws_specs().add_workers().parse_args()

    logging.info("AWS Data Downloading")

    location = Location(
        name=args.hemisphere,
        north=args.hemisphere == "north",
        south=args.hemisphere == "south",
    )

    dataset = AWSDatasetConfig(
        levels=args.levels,
        location=location,
        var_names=args.vars,
        frequency=getattr(Frequency, args.frequency),
        output_group_by=getattr(Frequency, args.output_group_by),
        config_path=args.config,
        overwrite=args.overwrite_config,
    )

    for start_date, end_date in zip(args.start_dates, args.end_dates):
        logging.info("Downloading between {} and {}".format(start_date, end_date))
        aws = AWSDownloader(
            dataset,
            start_date=start_date,
            end_date=end_date,
            delete_cache=args.delete_cache,
            cache_only=args.cache_only,
            compress=args.compress,
            max_threads=args.workers,
            request_frequency=getattr(Frequency, args.output_group_by),
        )
        aws.download()

        dataset.save_data_for_config(
            source_files=aws.files_downloaded,
            var_filter_list=["lambert_azimuthal_equal_area"],
        )



import logging
import os
import sys

import orjson

from download_toolbox.base import DataCollection
from download_toolbox.config import Configuration
from download_toolbox.dataset import DatasetConfig
from download_toolbox.location import Location
from download_toolbox.time import Frequency
from download_toolbox.utils import get_implementation

from download_toolbox.data.amsr import AMSRDatasetConfig
from download_toolbox.data.aws import AWSDatasetConfig
from download_toolbox.data.cds import ERA5DatasetConfig
from download_toolbox.data.esgf import CMIP6DatasetConfig
from download_toolbox.data.osisaf import SICDatasetConfig

__all__ = [
    "Configuration",
    "DataCollection",
    "DatasetConfig",
    "AMSRDatasetConfig",
    "AWSDatasetConfig",
    "ERA5DatasetConfig",
    "CMIP6DatasetConfig",
    "SICDatasetConfig",
    # Descriptions
    "Frequency",
    "Location",
    # Functions
    "get_dataset_config_implementation",
    "get_implementation",
]


def get_dataset_config_implementation(config: os.PathLike,
                                      output_path: os.PathLike = None):
    if not str(config).endswith(".json"):
        raise RuntimeError("{} does not look like a JSON configuration".format(config))
    if not os.path.exists(config):
        raise RuntimeError("{} is not a configuration in existence".format(config))

    logging.debug("Retrieving implementations details from {}".format(config))

    with open(config) as fh:
        data = fh.read()

    cfg = orjson.loads(data)
    logging.debug("Loaded configuration {}".format(",".join(cfg.keys())))
    cfg, implementation = cfg["data"], cfg["implementation"]

    # TODO: Getting a nicer implementation might be the way forward, but this will do
    #  with the Frequency naively matching given that they're fully caps-locked strings
    location = Location(**cfg["_location"])
    freq_dict = {k.strip("_"): getattr(Frequency, v) for k, v in cfg.items() if v in list(Frequency.__members__)}
    remaining = {k.strip("_"): v
                 for k, v in cfg.items()
                 if k not in [*["_{}".format(el) for el in freq_dict.keys()], "_location", "_config_type"]}

    create_kwargs = dict(config_path=output_path if output_path is not None else config,
                         location=location,
                         **remaining,
                         **freq_dict)
    logging.info("Attempting to instantiate {} with loaded configuration".format(implementation))
    logging.debug("Converted kwargs from the retrieved configuration: {}".format(",".join(create_kwargs.keys())))

    return get_implementation(implementation)(**create_kwargs)



import argparse
import datetime as dt
import logging
import re

from download_toolbox.time import Frequency

"""

"""


def date_arg(string: str) -> object:
    """

    :param string:
    :return:
    """
    date_match = re.search(r"(\d{4})-(\d{1,2})-(\d{1,2})", string)
    return dt.date(*[int(s) for s in date_match.groups()])


def dates_arg(string: str) -> object:
    """

    :param string:
    :return:
    """
    if string == "none":
        return []

    date_match = re.findall(r"(\d{4})-(\d{1,2})-(\d{1,2})", string)

    if len(date_match) < 1:
        raise argparse.ArgumentError(argument="dates",
                                     message="No dates found for supplied argument {}".format(string))
    return [dt.date(*[int(s) for s in date_tuple]) for date_tuple in date_match]


def csv_arg(string: str) -> list:
    """

    :param string:
    :return:
    """
    csv_items = []
    string = re.sub(r'^\'(.*)\'$', r'\1', string)

    for el in string.split(","):
        if len(el) == 0:
            csv_items.append(None)
        else:
            csv_items.append(el)
    return csv_items


def csv_of_csv_arg(string: str) -> list:
    """

    :param string:
    :return:
    """
    csv_items = []
    string = re.sub(r'^\'(.*)\'$', r'\1', string)

    for el in string.split(","):
        if len(el) == 0:
            csv_items.append(None)
        else:
            csv_items.append(el.split("|"))
    return csv_items


def csv_of_date_args(string: str) -> list:
    """

    :param string:
    :return:
    """
    csv_items = []
    string = re.sub(r'^\'(.*)\'$', r'\1', string)

    for el in string.split(","):
        if len(el) == 0:
            csv_items.append(None)
        else:
            csv_items.append([date_arg(date) for date in el.split("|")])
    return csv_items


def int_or_list_arg(string: str) -> object:
    """

    :param string:
    :return:
    """
    try:
        val = int(string)
    except ValueError:
        val = string.split(",")
    return val


def parse_known_args(unknown_args: list) -> dict:
    """
    Parses a list of unknown command line arguments and returns them.

    This function processes both long and short options, and supports boolean flags
    (e.g., `--flag` or `-f`). Non-option arguments are added to the dictionary as
    values associated with their preceding option keys.

    Args:
        unknown_args: A list of command line arguments.

    Returns:
        A dictionary containing parsed arguments.
        Keys are converted to snake_case for consistency,
        with how argparse would return them.
    """
    result = {}
    key = None

    for arg in unknown_args:
        if arg.startswith("--"):
            # Cover long option
            if key:
                result[key] = True
            key = arg.lstrip("-").replace("-", "_")
        elif len(arg) > 1 and arg.startswith("-"):
            # Cover short option
            key = arg.replace("-", "_")
            result[key] = True
        else:
            if key:
                result[key] = arg
                key = None
            else:
                pass

    if key:
        result[key] = True

    return result


class BaseArgParser(argparse.ArgumentParser):
    """An ArgumentParser specialised to support common argument handling

    The 'allow_*' methods return self to permit method chaining.

    :param suppress_logs:
    """

    def __init__(self,
                 *args,
                 log_format: str = "[%(asctime)-17s :%(levelname)-8s] - %(message)s",
                 suppress_logs: list = None,
                 **kwargs):
        super().__init__(*args, **kwargs)

        self._log_format = log_format
        self._suppress_logs = suppress_logs

        # FIXME: this is not ubiquitously useful in preprocess-toolbox: ref channel adds
        #   so there is a pattern failure having it at this level
        self.add_argument("-c", "--config-path",
                          dest="config",
                          help="Path at which to output the configuration when rendered")

        self.add_argument("-v",
                          "--verbose",
                          action="store_true",
                          default=False)

    def add_extra_args(self, extra_args):
        for arg in extra_args:
            self.add_argument(*arg[0], **arg[1])
        return self

    def parse_args(self,
                   *args,
                   **kwargs):
        args = super().parse_args(*args, **kwargs)

        self.setup_logging(verbose=args.verbose)

        return args

    def parse_known_args(self,
                         *args,
                         **kwargs):
        """
        Parses command line arguments and handles unknown arguments separately.

        This method first calls the argparse's `parse_known_args` method to separate
        known and unknown arguments. It then parses the unknown arguments using
        the `parse_known_args` function to convert the returned list to a dict.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            A tuple containing the parsed Namespace object for known arguments
            and a dictionary of parsed unknown arguments.
        """
        args, unknown_args_list = super().parse_known_args(*args, **kwargs)

        unknown_args = parse_known_args(unknown_args_list)

        self.setup_logging(verbose=args.verbose)

        return args, unknown_args

    def setup_logging(self, verbose: bool) -> None:
        """
        Configures logging based on verbosity and suppresses logs from specific modules.

        Args:
            verbose: Whether to enable debug-level logging.
        """
        # TODO: this is not necessarily ideal when running the argparser in notebooks
        loglevel = logging.DEBUG if verbose else logging.INFO
        logging.basicConfig(
            datefmt="%d-%m-%y %T",
            format=self._log_format,
            level=loglevel
        )
        logging.getLogger().setLevel(loglevel)

        if self._suppress_logs is not None and type(self._suppress_logs) is list:
            for log_module in self._suppress_logs:
                logging.debug("Setting {} to WARNING only".format(log_module))
                logging.getLogger(log_module).setLevel(logging.WARNING)

        # TODO: bring these out of defaults
        logging.getLogger("cdsapi").setLevel(logging.WARNING)
        logging.getLogger("requests").setLevel(logging.WARNING)
        logging.getLogger("urllib3").setLevel(logging.WARNING)
        logging.getLogger("matplotlib").setLevel(logging.WARNING)


class DownloadArgParser(BaseArgParser):
    def __init__(self,
                 *args,
                 dates_optional: bool = False,
                 **kwargs):
        super().__init__(*args, **kwargs)

        self._var_specs = False

        self.add_argument("hemisphere", choices=("north", "south"))

        date_arg_ids = [["start_dates"], ["end_dates"]] if not dates_optional else \
            [["-sd", "--start-dates"], ["-ed", "--end-dates"]]
        self.add_argument(*date_arg_ids[0], type=dates_arg, default=None)
        self.add_argument(*date_arg_ids[1], type=dates_arg, default=None)

        freq_avail = [_.name for _ in list(Frequency)]

        self.add_argument("-f", "--frequency",
                          choices=freq_avail,
                          default=freq_avail[-1])
        self.add_argument("-o", "--output-group-by",
                          choices=freq_avail,
                          default=freq_avail[0])

        self.add_argument("-oc", "--overwrite-config",
                          help="Overwrite dataset configuration",
                          action="store_true", default=False)

    def add_var_specs(self):
        self._var_specs = True

        self.add_argument("vars",
                          help="Comma separated list of vars",
                          type=csv_arg,
                          default=[])
        self.add_argument("levels",
                          help="Comma separated list of pressures/depths as needed, "
                               "use zero length string if None (e.g. ',,500,,,') and "
                               "pipes for multiple per var (e.g. ',,250|500,,'",
                          type=csv_of_csv_arg,
                          default=[])

        return self

    def add_workers(self):
        self.add_argument("-w", "--workers", default=8, type=int)
        return self

    def parse_args(self,
                   *args,
                   **kwargs):
        args = super().parse_args(*args, **kwargs)

        if self._var_specs:
            if not (len(args.vars) > 0 and len(args.vars) == len(args.levels)):
                raise RuntimeError("You must specify variables and levels of equal length, >=1: {} != {}".
                                   format(args.vars, args.levels))
        return args




from enum import Enum, auto


class Frequency(int, Enum):
    """

    https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#period-aliases

    """
    YEAR = 1, "%Y", "Y", "year", "yr", "%Y"
    MONTH = 2, "%Y%m", "M", "month", "mon", "%b %Y"
    DAY = 3, "%Y%m%d", "D", "day", "day", "%d %b %Y"
    HOUR = 4, "%Y%m%d%h", "H", "hour", "hr", "%d %b %Y, %I:00%p"

    def __new__(cls, value, date_format, freq, attribute, cmip_id, plot_format):
        member = int.__new__(cls, value)
        member._value_ = value
        member.date_format = date_format
        member.freq = freq
        member.attribute = attribute
        member.cmip_id = cmip_id
        member.plot_format = plot_format
        return member




from dataclasses import dataclass


@dataclass
class Location:
    """Representation of spatiotemporal location

    TODO: The intention is to converge on the geoapi representation

    https://www.geoapi.org/snapshot/python/metadata.html#spatial-representation
    """
    name: str
    bounds: tuple
    north: bool
    south: bool

    def __init__(self,
                 name: str,
                 bounds: tuple = None,
                 north: bool = False,
                 south: bool = False):
        self.name = name
        self.north = north
        self.south = south

        if bounds is not None and (north | south):
            raise RuntimeError("Provide a single location")

        self.bounds = list(bounds) if bounds is not None else \
            [90, -180, 0, 180] if north else \
            [0, -180, -90, 180] if south else \
            [90, -180, -90, 180]





import abc
import concurrent
import logging
from abc import ABCMeta, abstractmethod
from typing import Union

from concurrent.futures import ThreadPoolExecutor

import pandas as pd

from download_toolbox.data.utils import batch_requested_dates

from download_toolbox.dataset import DatasetConfig
from download_toolbox.time import Frequency

"""

"""


class Downloader(metaclass=abc.ABCMeta):
    """Abstract base class for a downloader.

    Performs operations on DataSets, we handle operations affecting the status of
    said DatasetConfig:
        1. Specify date range
        2. Specify batching behaviours:
            -

    """

    def __init__(self,
                 dataset: DatasetConfig,
                 *args,
                 delete_tempfiles: bool = True,
                 download: bool = True,
                 drop_vars: list = None,
                 end_date: list,
                 postprocess: bool = True,
                 request_frequency: object = Frequency.MONTH,
                 source_min_frequency: object = Frequency.DAY,
                 source_max_frequency: object = Frequency.DAY,
                 start_date: list,
                 **kwargs):
        super().__init__()

        self._dates = [pd.to_datetime(date).date() for date in
                       pd.date_range(start_date, end_date, freq=dataset.frequency.freq)]

        self._delete = delete_tempfiles
        self._download = download
        self._drop_vars = list() if drop_vars is None else drop_vars
        self._files_downloaded = []
        # TODO: can and should (?) be populated as part of download - threaded==future-arg map
        self._missing_dates = []
        self._postprocess = postprocess
        self._request_frequency = source_min_frequency \
            if request_frequency < source_min_frequency else source_max_frequency \
            if request_frequency > source_max_frequency else request_frequency
        self._source_min_frequency = source_min_frequency
        self._source_max_frequency = source_max_frequency

        logging.info("Request frequency set to {}".format(self.request_frequency.name))

        self._ds = dataset

        if not self._delete:
            logging.warning("!!! Deletions of temp files are switched off: be "
                            "careful with this, you need to manage your "
                            "files manually")

        self._download_method = self._single_download

    def download(self):
        """Implements a download for the given dataset

        This method handles download per var-"date batch" for the dataset
        """
        for var_config in self.dataset.variables:
            dates = self.dataset.filter_extant_data(var_config, self.dates)

            for req_date_batch in batch_requested_dates(dates=dates, attribute=self.request_frequency.attribute):
                logging.info("Processing download for {} with {} dates".
                             format(var_config.name, len(req_date_batch)))
                files_downloaded = self._download_method(var_config, req_date_batch)

                if files_downloaded is not None:
                    logging.info("{} files downloaded".format(len(files_downloaded)))
                    self._files_downloaded.extend(files_downloaded)
                else:
                    logging.warning("Nothing downloaded for {} on batch {}".format(var_config.name, req_date_batch))

    @abstractmethod
    def _single_download(self,
                         var_config: object,
                         req_dates: object) -> list:
        raise NotImplementedError("_single_download needs an implementation")

    @property
    def dataset(self):
        return self._ds

    @property
    def dates(self):
        return self._dates

    @property
    def delete(self):
        return self._delete

    @property
    def download_method(self) -> callable:
        if not self._download_method:
            raise RuntimeError("Downloader has no method set, "
                               "implementation error")
        return self._download_method

    @download_method.setter
    def download_method(self, method: callable):
        logging.debug("Setting download_method to {}".format(method))
        self._download_method = method

    @property
    def drop_vars(self):
        return self._drop_vars

    @property
    def files_downloaded(self):
        return self._files_downloaded

    @property
    def missing_dates(self):
        return self._missing_dates

    @property
    def request_frequency(self) -> Frequency:
        return self._request_frequency

    @property
    def skipped_dates(self) -> set:
        return self._skipped_dates


class ThreadedDownloader(Downloader, metaclass=ABCMeta):
    """Data downloader base class for batching downloading

    :param dates:
    :param delete_tempfiles:
    :param download:
    :param group_dates_by:
    :param max_threads:
    :param postprocess:
    :param var_name_idx:
    """

    def __init__(self, *args,
                 max_threads: int = 1,
                 **kwargs):
        super().__init__(*args, **kwargs)

        self._max_threads = max_threads

    def download(self):
        """Handles concurrent (threaded) downloading for variables

        This takes dates, variables and levels as configured, batches them into
        requests and submits those via a ThreadPoolExecutor for concurrent
        downloading. Returns nothing, relies on _single_download to implement
        appropriate updates to this object to record state changes arising from
        downloading.
        """

        logging.info("Building request(s), downloading and averaging "
                     "from {} API".format(self.dataset.identifier.upper()))

        req_list = list()

        for var_config in self.dataset.variables:
            dates = self.dataset.filter_extant_data(var_config, self.dates)

            for req_date_batch in batch_requested_dates(dates=dates, attribute=self.request_frequency.attribute):
                logging.info("Processing single download for {} with {} dates".
                             format(var_config.name, len(req_date_batch)))

                req_list.append((var_config, req_date_batch))

        max_workers = min(len(req_list), self._max_threads)

        if max_workers > 0:
            logging.info("Creating thread pool with {} workers to service {} batches"
                         .format(max_workers, len(req_list)))

            with ThreadPoolExecutor(max_workers=max_workers) as executor:
                futures = []

                for args in req_list:
                    future = executor.submit(self.download_method, *args)
                    futures.append(future)

                for future in concurrent.futures.as_completed(futures):
                    try:
                        files_downloaded = future.result()

                        if files_downloaded is not None:
                            logging.info("{} files downloaded".format(len(files_downloaded)))
                            self._files_downloaded.extend(files_downloaded)
                        else:
                            logging.warning("Nothing downloaded from threaded batch")

                    except Exception as e:
                        logging.exception("Thread failure: {}".format(e))

        logging.info("{} files downloaded".format(len(self._files_downloaded)))

    @property
    def max_threads(self):
        return self._max_threads


class DownloaderError(RuntimeError):
    pass



import json
import logging
import os
import shutil
import tempfile
from dataclasses import dataclass
from pprint import pformat

import dask
import pandas as pd
import xarray as xr

from download_toolbox.base import DataCollection
from download_toolbox.config import Configuration
from download_toolbox.data.utils import merge_files
from download_toolbox.time import Frequency


@dataclass
class VarConfig:
    name: str
    prefix: str
    level: int
    path: os.PathLike
    root_path: os.PathLike


class DatasetConfig(DataCollection):
    """A datasetconfig is an implementation of the base data collection, adding characteristics.

    Yes, this is intentionally not called a dataset as it doesn't override xarray.Dataset and
    it feels nicer that it represents a configuration for a Dataset, rather than a Dataset itself

    The additional characteristics that are implemented at this level:
        1. Location awareness
        2. Variables and levels

    TODO: align to https://www.geoapi.org/snapshot/python/metadata.html#metadata-iso-19115
      - intention is to eventually describe the DatasetConfig and metadata conformantly

    :param overwrite: Flag specifying whether existing files should be overwritten or not.
    """

    def __init__(self,
                 *,
                 config_type: str = "dataset_config",
                 existing_dates: list = None,
                 frequency: object = Frequency.DAY,
                 levels: object = (),
                 location: object,
                 output_group_by: object = Frequency.YEAR,
                 overwrite: bool = False,
                 path_components: list = None,
                 # TODO: Perhaps review the implementation with Enum to a bitwise typed one @ Py3.9+
                 valid_frequencies: tuple = (Frequency.DAY, Frequency.MONTH),
                 var_files: dict = None,
                 var_names: object = (),
                 **kwargs) -> None:
        super(DatasetConfig, self).__init__(config_type=config_type,
                                            path_components=[frequency.name.lower(), location.name]
                                            if path_components is None else path_components,
                                            **kwargs)

        self._existing_dates = list() if existing_dates is None else existing_dates
        self._frequency = frequency
        self._levels = list(levels)
        self._location = location
        self._output_group_by = output_group_by
        self._overwrite = overwrite
        self._var_files = dict() if var_files is None else var_files
        self._var_names = list(var_names)

        if len(self._var_names) < 1:
            raise DataSetError("No variables requested")

        if len(self._levels) != len(self._var_names):
            raise DataSetError("# of levels must match # vars")

        if self._frequency < self._output_group_by:
            raise DataSetError("You can't request a higher output frequency than request frequency: {} vs {}".
                               format(self._output_group_by.name, self._frequency.name))

        if self._frequency not in valid_frequencies:
            raise DataSetError("Only the following frequencies are valid for request: {}".format(valid_frequencies))

    def _get_data_var_folder(self,
                             var: str,
                             root: bool = False,
                             append: object = None,
                             missing_error: bool = False) -> str:
        """Returns the path for a specific data variable.

        Appends additional folders to the path if specified in the `append` parameter.

        :param var: The data variable.
        :param append: Additional folders to append to the path. Defaults to None.
        :param missing_error: Flag to specify if missing directories should be treated as an error. Defaults to False.
        :returns str: The path for the specific data variable.
        """
        if not append:
            append = []

        data_var_path = os.path.join(self.path if not root else self.root_path, *[var, *append])

        if not os.path.exists(data_var_path):
            if not missing_error:
                os.makedirs(data_var_path, exist_ok=True)
            else:
                raise OSError("Directory {} is missing and this is "
                              "flagged as an error!".format(data_var_path))

        return data_var_path

    def copy_to(self,
                new_identifier: object,
                base_path: os.PathLike = None,
                skip_copy: bool = False) -> object:
        """

        Args:
            new_identifier:
            base_path:
            skip_copy:
        """
        old_path = self.path
        super().copy_to(new_identifier, base_path, skip_copy=True)
        logging.info("Applying copy_to to identifier {}".format(new_identifier))

        for var_name in self.var_files.keys():
            old_files = self.var_files[var_name]
            new_files = [var_file.replace(old_path, self.path) for var_file in old_files]
            invalid_files = []

            for src, dest in zip(old_files, new_files):
                if os.path.exists(src):
                    logging.debug("Copying {} to {}".format(src, dest))
                    os.makedirs(os.path.dirname(dest), exist_ok=True)
                    shutil.copy(src, dest)
                else:
                    logging.warning("Encountered reference to non-existent data: {}".format(src))
                    invalid_files.append(dest)
            self.var_files[var_name] = [fn for fn in new_files if fn not in invalid_files]

    def filter_extant_data(self,
                           var_config: VarConfig,
                           dates: list) -> list:
        dt_arr = list(reversed(sorted(dates)))
        filepaths = self.var_filepaths(var_config, dt_arr)

        # Filtering dates based on existing data
        extant_paths = set([filepath
                            for filepath in filepaths
                            if os.path.exists(filepath)])
        logging.info("Filtering {} dates against {} destination files".format(len(dt_arr), len(filepaths)))
        logging.debug("Filtering against: {}".format(pformat(filepaths)))

        if len(extant_paths) > 0:
            extant_ds = xr.open_mfdataset(extant_paths)
            # This is blunt initialisation, as we're completely refreshing from source
            self._existing_dates = [pd.to_datetime(d).date()
                                    for d in extant_ds.time.values]

            dt_arr = sorted(list(set(dt_arr).difference(self._existing_dates)))
            dt_arr.reverse()

            # We won't hold onto an active dataset during network I/O
            extant_ds.close()
            logging.debug("{} dates filtered down to {} dates".format(len(dates), len(dt_arr)))
        return dt_arr

    def get_config(self,
                   config_funcs: dict = None,
                   strip_keys: list = None):
        my_keys = ["_overwrite"]

        def merge_var_files(x):
            data = dict() \
                if ("_var_files" not in self.config.data
                    or self.config.data["_var_files"] is None) \
                else self.config.data["_var_files"].copy()

            for var_name in x.keys():
                if var_name not in data:
                    data[var_name] = list()

                if isinstance(x[var_name], str):
                    data[var_name].append(x[var_name])
                else:
                    data[var_name].extend(x[var_name])
            return {k: list(sorted(set(files))) for k, files in data.items()}

        my_funcs = dict(
            _frequency=lambda x: x.name,
            _location=lambda x: dict(name=x.name, bounds=x.bounds)
            if not x.north and not x.south else dict(name=x.name, north=x.north, south=x.south),
            _output_group_by=lambda x: x.name,
            # TODO: this can't be done like this as the levels and var_names are ordered - GH#51
            # _var_names=lambda x: self._var_names + x,
            _var_files=merge_var_files
        )

        config_funcs = {} if config_funcs is None else config_funcs
        strip_keys = my_keys if strip_keys is None else my_keys + strip_keys
        return super().get_config(config_funcs={**my_funcs, **config_funcs},
                                  strip_keys=strip_keys)

    def get_dataset(self,
                    var_names: list = None):
        if var_names is None:
            logging.debug(self.variables)
            var_names = [v.name for v in self.variables]

        logging.debug("Finding files for {}".format(", ".join(var_names)))
        var_files = [var_filepath
                     for vn in var_names
                     for var_filepath in self.var_files[vn]
                     if os.path.exists(var_filepath)]
        logging.info("Got {} filenames to open dataset with!".format(len(var_files)))

        # TODO: where's my parallel mfdataset please!?
        with dask.config.set(**{'array.slicing.split_large_chunks': True,
                                # "scheduler": self._scheduler, # Fix to "single-threaded" for netCDF4 >=1.6.1 not thread-safe.
                                }):
            ds = xr.open_mfdataset(
                var_files,
                combine="nested",
                concat_dim="time",
                coords="minimal",
                compat="override"
            )

            ds = ds.drop_duplicates("time").chunk(dict(time=1, ))
        return ds

    def save_data_for_config(self,
                             combine_method: str = "by_coords",
                             rename_var_list: dict = None,
                             source_ds: object = None,
                             source_files: list = None,
                             time_dim_values: list = None,
                             var_filter_list: list = None):
        # Check whether we have a valid source
        ds = None
        if type(source_ds) in [xr.Dataset, xr.DataArray]:
            ds = source_ds if type(source_ds) is xr.Dataset else source_ds.to_dataset()

            if source_files is not None:
                raise RuntimeError("Not able to combine sources in save_dataset at present")
        elif source_files is not None and len(source_files) > 0:
            try:
                logging.debug("Opening source files: {}".format(pformat(source_files)))
                ds = xr.open_mfdataset(source_files,
                                       combine=combine_method,
                                       concat_dim=None if combine_method == "by_coords" else "time",
                                       parallel=True,
                                       engine="h5netcdf",
                                       lock=False,      # Attempt to avoid deadlocks, xarray GH#3961
                                       )
            except ValueError as e:
                logging.exception("Could not open files {} with error".format(source_files))
                raise DataSetError(e)

            if time_dim_values is not None:
                time_dt_arr = [pd.Timestamp(d) for d in time_dim_values if d not in self._existing_dates]
                logging.warning("Assigning time dimension with {} values".format(len(time_dt_arr)))
                ds = ds.assign(dict(time=time_dt_arr))
                self._existing_dates.append([pd.to_datetime(d).date() for d in time_dt_arr])
        else:
            logging.warning("No data provided as data object or source files, not doing anything")
            if self._overwrite:
                logging.warning("Overwriting configuration even without data thanks to dataset.overwrite flag")
                self.save_config()
            return

        # Strip out unnecessary / unwanted variables
        if var_filter_list is not None:
            ds = ds.drop_vars(var_filter_list, errors="ignore")

        # TODO: Reduce spatially to required location this will also need to set our shape details
        # TODO: ideally we should have a broader cache that would allow us to reuse data
        # TODO: we CANNOT handle nav_lon and nav_lat (e.g. 2D point meshes) yet
        #if all([f in ds.coords for f in ["latitude", "longitude"]]):
        #    ds = ds.sel(latitude=slice(self.location.bounds[0], self.location.bounds[2]),
        #                longitude=slice(self.location.bounds[1], self.location.bounds[3]))

        # Reduce temporally to required resolution
        # TODO: Note, https://github.com/pydata/xarray/issues/364 for Grouper functionality?
        #   - we might have to roll our own functionality in the meantime, if necessary
        group_by = "time.{}".format(self._output_group_by.attribute)

        # Rename if requested
        if rename_var_list:
            logging.info("Renaming {} variables if available".format(rename_var_list))
            ds = ds.rename({k: v for k, v in rename_var_list.items() if k in list(ds.coords) + list(ds.data_vars)})

        # For all variables in ds, determine if there are destinations available
        for var_config in [vc for vc in self.variables if vc.name in ds.data_vars]:
            da = getattr(ds, var_config.name)
            logging.debug("Resampling to period 1{}: {}".format(self.frequency.freq, da.shape))
            da = da.sortby("time").resample(time="1{}".format(self.frequency.freq)).mean(keep_attrs=True)

            logging.debug("Grouping {} by {}".format(var_config, group_by))
            for dt, dt_da in da.groupby(group_by):
                req_dates = pd.to_datetime(dt_da.time.values)
                logging.debug("Have group of {} dates".format(len(req_dates)))
                destination_path = self.var_filepath(var_config, req_dates)

                copy_attrs = {k: v for k, v in ds.attrs.items() if k.startswith("geospatial")}
                logging.debug("Reassinging geospatial info to derived dataset: {}".format(copy_attrs))
                dt_ds = dt_da.to_dataset().assign_attrs(copy_attrs)

                # If exists, merge and concatenate the data to destination (overwrite?) at output_group_by
                if os.path.exists(destination_path):
                    logging.debug("Outputting new data to temporary file as {} already exists".format(destination_path))
                    fh, temporary_name = tempfile.mkstemp(dir=".")
                    os.close(fh)
                    dt_ds.to_netcdf(temporary_name)
                    dt_ds.close()
                    logging.info("Written new data to {} and merging with {}".format(
                        temporary_name, destination_path
                    ))
                    merge_files(destination_path, temporary_name)
                else:
                    logging.info("Saving {}".format(destination_path))
                    dt_ds.to_netcdf(destination_path)
                    dt_ds.close()

        # Write out the configuration file
        self.save_config()

    def var_config(self, var_name, level=None):
        """

        :param var_name:
        :param level:
        :return:
        """
        var_full_name = "{}{}".format(var_name,
                                      str(level) if level is not None else "")

        return VarConfig(
            name=var_full_name,
            prefix=var_name,
            level=level,
            path=self._get_data_var_folder(var_full_name),
            root_path=self._get_data_var_folder(var_full_name, root=True)
        )

    def var_filepath(self, *args, **kwargs) -> os.PathLike:
        return self.var_filepaths(*args, **kwargs, single_only=True)[0]

    def var_filepaths(self,
                      var_config: VarConfig,
                      date_batch: list,
                      file_extension: str = "nc",
                      single_only: bool = False) -> list:
        """

        :param var_config:
        :param date_batch:
        :param single_only:
        :return:
        """
        output_filepaths = list(set([
            os.path.join(var_config.path, "{}.{}".format(
                date.strftime(self._output_group_by.date_format),
                file_extension
            ))
            for date in date_batch]))

        if len(output_filepaths) > 1 and single_only:
            raise DataSetError("Filenames returned for {} dates should have been "
                               "singular but {} returnable, check your call / config".
                               format(len(date_batch), len(output_filepaths)))

        if len(output_filepaths) == 0:
            logging.warning("No filenames provided for {} - {}".format(var_config, len(date_batch)))

        self._var_files[var_config.name] = list(set(
            self._var_files[var_config.name] + output_filepaths
        )) if var_config.name in self._var_files else output_filepaths

        return output_filepaths

    @property
    def frequency(self):
        return self._frequency

    @property
    def location(self):
        return self._location

    @property
    def variables(self):
        for var_name, levels in zip(self._var_names, self._levels):
            for level in levels if levels is not None else [None]:
                var_config = self.var_config(var_name, level)
                yield var_config

    @property
    def var_files(self):
        return self._var_files

    @property
    def var_prefixes(self):
        return self._var_names

    @var_files.setter
    def var_files(self, value: dict):
        logging.warning("Setting new file setup to dataset with {} files".format(
            ", ".join(["{} for {}".format(len(v), k) for k, v in value.items()])))
        self._var_files = value

    def __repr__(self):
        return pformat(self.__dict__)


class DataSetError(RuntimeError):
    pass




import logging
import os

import datetime as dt

from download_toolbox.dataset import DatasetConfig, DataSetError
from download_toolbox.cli import DownloadArgParser
from download_toolbox.download import ThreadedDownloader, DownloaderError
from download_toolbox.utils import HTTPClient, ClientError, DaskWrapper
from download_toolbox.location import Location
from download_toolbox.time import Frequency


var_remove_list = []  # ["polar_stereographic", "land"]


class AMSRDatasetConfig(DatasetConfig):
    def __init__(self,
                 identifier=None,
                 levels=None,
                 resolution=6.25,
                 var_names=None,
                 **kwargs):
        # There are other resolutions available, this will need updating for more
        # products, such as 1km AMSR+MODIS
        if resolution not in (3.125, 6.25):
            raise DataSetError("{} is not a valid resolution".format(resolution))

        self._resolution = resolution

        super().__init__(identifier="amsr2_{:1.3f}".format(resolution).replace(".", "")
                         if identifier is None else identifier,
                         # TODO: see below, GH#10 on consistency of naming in datasets, this should be in IceNet
                         var_names=["siconca"] if var_names is None else var_names,
                         levels=[None] if levels is None else levels,
                         **kwargs)

    @property
    def resolution(self):
        return self._resolution


class AMSRDownloader(ThreadedDownloader):
    """Downloads AMSR2 SIC data from 2012-present using HTTPS.

    The data can come from yearly zips, or individual files. We target the individual files as in reality it's so much
    more sensible - the zips are not consistently avialable across the data ranges

    We use the following for HTTPS downloads:
        - data.seaice.uni-bremen.de
    But can also use (noticed when the above was down!)
        - seaice.uni-bremen.de/data/

    """
    def __init__(self,
                 dataset: AMSRDatasetConfig,
                 *args,
                 start_date: object,
                 host: str = "data.seaice.uni-bremen.de",
                 **kwargs):
        amsr2_start = dt.date(2012, 7, 2)

        # TODO: Differing start date ranges for different products! Validate in dataset
        if start_date < amsr2_start:
            raise DownloaderError("AMSR2 only exists past {}".format(amsr2_start))
        self._hemi_str = "s" if dataset.location.south else "n"

        self._http_client = HTTPClient("https://{}".format(host),
                                       source_base="amsr2/asi_daygrid_swath/{}{}/netcdf".format(
                                           self._hemi_str, "{:1.3f}".format(dataset.resolution).replace(".", "")))

        super().__init__(dataset,
                         *args,
                         start_date=start_date,
                         **kwargs)

    def _single_download(self,
                         var_config: object,
                         req_dates: object):

        files_downloaded = []

        for file_date in req_dates:
            year_dir = str(file_date.year)
            date_str = file_date.strftime("%Y%m%d")

            file_in_question = "{}/asi-AMSR2-{}{}-{}-v5.4.nc".\
                               format(year_dir, self._hemi_str, "{:1.3f}".
                                      format(self.dataset.resolution).
                                      replace(".", ""), date_str)
            destination_path = os.path.join(var_config.root_path, file_in_question)

            if not os.path.exists(os.path.dirname(destination_path)):
                os.makedirs(os.path.dirname(destination_path), exist_ok=True)

            if not os.path.exists(destination_path):
                try:
                    logging.info("Downloading {}".format(destination_path))
                    self._http_client.single_request(file_in_question, destination_path)
                    files_downloaded.append(destination_path)
                except (ClientError, DownloaderError) as e:
                    logging.warning("Failed to download {}: {}".format(destination_path, e))
                    self.missing_dates.append(file_date)
            else:
                logging.debug("{} already exists".format(destination_path))
                files_downloaded.append(destination_path)

        return files_downloaded


def main():
    args = DownloadArgParser().add_workers().add_extra_args([
        (["--host"], dict(
            type=str,
            default="data.seaice.uni-bremen.de",
            help="Host portion of URI to download from",
        )),
        (["-r", "--resolution"], dict(
            type=float,
            choices=[3.125, 6.25],
            default=6.25
        )),
        (["-dw", "--dask-workers"], dict(
            type=int,
            default=4,
            help="Number of dask workers for saving files"
        ))
    ]).parse_args()

    logging.info("AMSR-SIC Data Downloading")
    location = Location(
        name=args.hemisphere,
        north=args.hemisphere == "north",
        south=args.hemisphere == "south",
    )

    dataset = AMSRDatasetConfig(
        location=location,
        frequency=getattr(Frequency, args.frequency),
        output_group_by=getattr(Frequency, args.output_group_by),
        config_path=args.config,
        overwrite=args.overwrite_config,
    )

    for start_date, end_date in zip(args.start_dates, args.end_dates):
        logging.info("Downloading between {} and {}".format(start_date, end_date))
        sic = AMSRDownloader(
            dataset,
            host=args.host,
            max_threads=args.workers,
            start_date=start_date,
            end_date=end_date,
        )
        sic.download()

        # TODO: DaskWrapper is a bit dirty, but works for overcoming some issues around processing speed
        with DaskWrapper(workers=args.dask_workers):
            dataset.save_data_for_config(
                combine_method="nested",
                # TODO: This should ideally be in IceNet? There is a bigger issue of naming to address (GH#10)
                rename_var_list=dict(z="siconca"),
                source_files=sic.files_downloaded,
                time_dim_values=[date for date in sic.dates if date not in sic.missing_dates],
                var_filter_list=var_remove_list)



import logging
import os
import requests
import urllib

from functools import lru_cache

import datetime as dt
import xarray as xr

from download_toolbox.dataset import DatasetConfig, DataSetError
from download_toolbox.cli import DownloadArgParser
from download_toolbox.download import ThreadedDownloader, DownloaderError
from download_toolbox.location import Location
from download_toolbox.time import Frequency

"""

"""


class CMIP6DatasetConfig(DatasetConfig):

    DAY_TABLE_MAP = {
        'siconc': 'SI{}',
        'siconca': 'SI{}',
        'tas': '{}',
        'ta': '{}',
        'tos': 'O{}',
        'hus': '{}',
        'psl': '{}',
        'rlds': '{}',
        'rsus': '{}',
        'rsds': '{}',
        'zg': '{}',
        'uas': '{}',
        'vas': '{}',
        'ua': '{}',
    }

    MONTH_TABLE_MAP = {
        'siconc': 'SI{}',
        'siconca': 'SI{}',
        'tas': 'A{}',
        'ta': 'A{}',
        'tos': 'O{}',
        'hus': 'A{}',
        'psl': 'A{}',
        'rlds': 'A{}',
        'rsus': 'A{}',
        'rsds': 'A{}',
        'zg': 'A{}',
        'uas': 'A{}',
        'vas': 'A{}',
        'ua': 'A{}'
    }

    GRID_MAP = {
        "EC-Earth3": {
            'siconc': 'gr',
            'siconca': 'gr',
            'tas': 'gr',
            'ta': 'gr',
            'tos': 'gr',
            'hus': 'gr',
            'psl': 'gr',
            'rlds': 'gr',
            'rsus': 'gr',
            'rsds': 'gr',
            'zg': 'gr',
            'uas': 'gr',
            'vas': 'gr',
            'ua': 'gr',
        },
        "MRI-ESM2-0": {
            'siconc': 'gn',
            'siconca': 'gn',
            'tas': 'gn',
            'ta': 'gn',
            'tos': 'gr',
            'hus': 'gn',
            'psl': 'gn',
            'rlds': 'gn',
            'rsus': 'gn',
            'rsds': 'gn',
            'zg': 'gn',
            'uas': 'gn',
            'vas': 'gn',
            'ua': 'gn',
        }
    }

    def __init__(self,
                 source: str,
                 member: str,
                 experiments: object = (
                    "historical",
                    "ssp245",
                 ),
                 default_grid: object = None,
                 grid_override: object = None,
                 identifier=None,
                 table_map_override: object = None,
                 **kwargs):
        super().__init__(identifier="cmip6.{}.{}".format(source, member)
                         if identifier is None else identifier,
                         **kwargs)

        self._experiments = experiments
        self._member = member
        self._source = source

        self._grid_override = dict() if grid_override is None else grid_override
        self._table_map_override = dict() if table_map_override is None else table_map_override

        if type(self._grid_override) is not dict:
            raise DataSetError("Grid override should be a dictionary if supplied")
        if type(self._table_map_override) is not dict:
            raise DataSetError("Table map override should be a dictionary if supplied")

        self._grid_map = CMIP6DatasetConfig.GRID_MAP
        if default_grid is not None:
            self._grid_map = {k: default_grid for k in CMIP6DatasetConfig.GRID_MAP.keys()}
        self._grid_map.update(self._grid_override)

        self._table_map = {k: v.format(self.frequency.cmip_id) for k, v in
                           getattr(CMIP6DatasetConfig, "{}_TABLE_MAP".format(self.frequency.name.upper())).items()}

    @property
    def experiments(self):
        return self._experiments

    @property
    def grid_map(self):
        return self._grid_map

    @property
    def member(self):
        return self._member

    @property
    def source(self):
        return self._source

    @property
    def table_map(self):
        return self._table_map


class CMIP6LegacyDownloader(ThreadedDownloader):
    # Prioritise European first, US last, avoiding unnecessary queries
    # against nodes further afield (all traffic has a cost, and the coverage
    # of local nodes is more than enough)
    def __init__(self,
                 *args,
                 nodes: object = None,
                 exclude_nodes: object = None,
                 search_node: object = "https://esgf-node.llnl.gov/esg-search/search",
                 **kwargs):
        super().__init__(*args,
                         # TODO: validate ESGF frequencies other than MONTH / YEAR
                         source_min_frequency=Frequency.YEAR,
                         source_max_frequency=Frequency.HOUR,
                         # TODO: request_frequency is a problem, we should batch
                         #  as big as possible, but currently only cache queries
                         **kwargs)
        exclude_nodes = list() if exclude_nodes is None else exclude_nodes

        self._search_node = search_node
        # self.__connection = None
        self._exclude_nodes = exclude_nodes if exclude_nodes is not None else []

    def _single_download(self,
                         var_config: object,
                         req_dates: object) -> list:
        """Overridden CMIP implementation for downloading from DAP server

        Due to the size of the CMIP set and the fact that we don't want to make
        1850-2100 yearly requests for all downloads, we have a bespoke and
        overridden download implementation for this.

        TODO: this could be made to handle much larger date ranges, as individual
         files contain centuries of data. Making a search / download per year isn't ideal

        :param var_config:
        :param req_dates:
        """

        query = {
            'source_id': self.dataset.source,
            'member_id': self.dataset.member,
            'frequency': self.dataset.frequency.cmip_id,
            'variable_id': var_config.prefix,
            'table_id': self.dataset.table_map[var_config.prefix],
            'grid_label': self.dataset.grid_map[self.dataset.source][var_config.prefix],
        }

        results = []

        logging.info("Querying ESGF for experiment {} for {}".format(" and ".join(self.dataset.experiments), var_config.name))
        query['experiment_id'] = ",".join(self.dataset.experiments)
        node_results = None

        try:
            node_results = self.esgf_search(**query)
        except requests.exceptions.HTTPError as e:
            logging.error(f"HTTP error {e.response} with collecting node results, "
                          f"clearing the cache and maybe try a different search node")
            self.esgf_search.cache_clear()

        if node_results is not None and len(node_results) > 0:
            logging.debug("Query: {}".format(query))
            logging.debug("Found {}-{}: {} results".format(
                ",".join(self.dataset.experiments), var_config.name, len(node_results)))

            start_date, end_date = req_dates[0], req_dates[-1]
            date_proc = lambda s: dt.datetime(int(s[0:4]), int(s[4:6]), 1).date() \
                if self.dataset.frequency <= Frequency.MONTH else dt.datetime(int(s[0:4]), int(s[4:6]),
                                                                              int(s[6:8])).date()
            for idx, df in enumerate(node_results):
                start, end = [date_proc(date_str) for date_str in df.split("_")[-1].rstrip(".nc").split("-")]
                if start <= start_date < end_date <= end \
                        or start <= start_date <= end \
                        or start <= end_date <= end:
                    results.append(df)
        else:
            logging.warning("No unbounded results for {} from ESGF search {}".format(
                var_config.name, ",".join(query.values())))
            return None

        # Filter by excluded domains and group together the rest for testing
        node_grouped_results = dict()
        for result in results:
            host = urllib.parse.urlparse(result).hostname
            if any([excl in host for excl in self._exclude_nodes]):
                logging.debug("Skipping {} as in the excluded hosts list".format(host))
                continue
            if host not in node_grouped_results:
                node_grouped_results[host] = []
            node_grouped_results[host].append(result)

        if len(node_grouped_results) == 0:
            # TODO: what really happens when we have this?
            logging.warning("NO VALID URLs FOUND for {} from ESGF search {}".format(
                var_config.name, ",".join(query.values())))
            return None

        cmip6_ds = None
        download_path = os.path.join(var_config.root_path,
                                     self.dataset.location.name,
                                     os.path.basename(self.dataset.var_filepath(var_config, req_dates)))

        if os.path.exists(download_path):
            logging.warning(f"We have downloaded data without corresponding output, "
                            f"so will skip the download: {download_path}")
            return [download_path]

        logging.debug("\n".join(results))

        for node, grouped_results in node_grouped_results.items():
            logging.info("Attempting to open data from {}".format(node))
            try:
                # http://xarray.pydata.org/en/stable/user-guide/io.html?highlight=opendap#opendap
                # Avoid 500MB DAP request limit
                cmip6_ds = xr.open_mfdataset(grouped_results,
                                             combine='by_coords',
                                             chunks={'time': '499MB'})
            except OSError as e:
                logging.error("Could not open data from {} - {}".format(node, e.filename))
                continue
            break

        if cmip6_ds is None:
            logging.error("We have no URLS that provide data for {} using {}".format(var_config.name, ",".join(query.values())))
            return None

        try:
            rename_vars = {var_config.prefix: var_config.name}
            cmip6_da = getattr(cmip6_ds.rename(rename_vars), var_config.name)

            if self.dataset.frequency == Frequency.MONTH and start_date.day > 1:
                start_date = start_date.replace(day=1)

            cmip6_da = cmip6_da.sel(time=slice(start_date,
                                               end_date))

            # TODO: possibly other attributes, especially with ocean vars
            if var_config.level:
                cmip6_da = cmip6_da.sel(plev=int(var_config.level) * 100)

            # Note this is happening on the remote end!
            cmip6_da = cmip6_da.sel(lat=slice(self.dataset.location.bounds[2],
                                              self.dataset.location.bounds[0]))

            # By this point the variable name has stored this info
            for omit_coord in ["plev", "height"]:
                if omit_coord in cmip6_da.coords:
                    cmip6_da = cmip6_da.drop_vars(omit_coord)
        except (OSError, ValueError, IndexError) as e:
            raise DownloaderError("Error encountered: {} for {}".format(e, results))
        else:
            logging.info("Writing {} to {}".format(cmip6_da, download_path))
            os.makedirs(os.path.dirname(download_path), exist_ok=True)
            cmip6_da.to_netcdf(download_path)
            cmip6_da.close()

        if os.path.exists(download_path):
            return [download_path]

        return None

    @lru_cache(1000)
    def esgf_search(self,
                    files_type: str = "OPENDAP",
                    local_node: bool = False,
                    latest: bool = True,
                    project: str = "CMIP6",
                    format: str = "application%2Fsolr%2Bjson",
                    use_csrf: bool = False,
                    **search):
        """

        Below taken from
        https://hub.binder.pangeo.io/user/pangeo-data-pan--cmip6-examples-ro965nih/lab
        and adapted slightly

        :param files_type:
        :param local_node:
        :param latest:
        :param project:
        :param format:
        :param use_csrf:
        :param search:
        :return:
        """
        client = requests.session()
        payload = search
        payload["project"] = project
        payload["type"] = "File"
        if latest:
            payload["latest"] = "true"
        if local_node:
            payload["distrib"] = "false"
        if use_csrf:
            client.get(self._search_node)
            if 'csrftoken' in client.cookies:
                # Django 1.6 and up
                csrftoken = client.cookies['csrftoken']
            else:
                # older versions
                csrftoken = client.cookies['csrf']
            payload["csrfmiddlewaretoken"] = csrftoken

        payload["format"] = format

        offset = 0
        numFound = 10000
        all_files = []
        files_type = files_type.upper()
        payload["limit"] = 1000
        while offset < numFound:
            payload["offset"] = offset
            url_keys = []
            for k in payload:
                url_keys += ["{}={}".format(k, payload[k])]

            url = "{}/?{}".format(self._search_node, "&".join(url_keys))
            logging.debug("ESGF search URL: {}".format(url))

            r = client.get(url)
            r.raise_for_status()
            resp = r.json()["response"]
            numFound = int(resp["numFound"])
            resp = resp["docs"]
            offset += len(resp)
            for d in resp:
                # for k in d:
                #    logging.debug("{}: {}".format(k, d[k]))

                for f in d["url"]:
                    sp = f.split("|")
                    if sp[-1] == files_type:
                        all_files.append(sp[0].split(".html")[0])

        return sorted(all_files)


def main():
    args = DownloadArgParser().add_var_specs().add_workers().add_extra_args([
        (["--source"], dict(type=str, default="MRI-ESM2-0")),
        (["--member"], dict(type=str, default="r1i1p1f1")),
        (("-n", "--search-node"), dict(type=str, default="https://esgf-node.llnl.gov/esg-search/search")),
        # (["--pyesgf"], dict(default=False, action="store_true")),
        (("-xs", "--exclude-server"),
         dict(default=[], nargs="*")),
        # (("-o", "--grid-override"), dict(required=None, type=str)),
        (("-g", "--default-grid"), dict(required=None, type=str)),
    ]).parse_args()

    logging.info("CMIP6 Data Downloading")

    location = Location(
        name=args.hemisphere,
        north=args.hemisphere == "north",
        south=args.hemisphere == "south",
    )

    dataset = CMIP6DatasetConfig(
        levels=args.levels,
        location=location,
        member=args.member,
        source=args.source,
        var_names=args.vars,
        frequency=getattr(Frequency, args.frequency),
        output_group_by=getattr(Frequency, args.output_group_by),
        config_path=args.config,
        overwrite=args.overwrite_config,
    )

    for start_date, end_date in zip(args.start_dates, args.end_dates):
        logging.info("Downloading between {} and {}".format(start_date, end_date))
        cmip6 = CMIP6LegacyDownloader(
            dataset=dataset,
            start_date=start_date,
            end_date=end_date,
            max_threads=args.workers,
            exclude_nodes=args.exclude_server,
            request_frequency=getattr(Frequency, args.output_group_by),
            search_node=args.search_node,
        )

        logging.info("CMIP downloading: {} {}".format(args.source, args.member))
        cmip6.download()
        dataset.save_data_for_config(
            source_files=cmip6.files_downloaded,
        )



import logging
import os

from typing import Union

import ecmwfapi
import xarray as xr

from download_toolbox.cli import DownloadArgParser, csv_arg
from download_toolbox.dataset import DatasetConfig
from download_toolbox.data.utils import xr_save_netcdf
from download_toolbox.download import Downloader, ThreadedDownloader, DownloaderError
from download_toolbox.location import Location
from download_toolbox.time import Frequency


class MARSDownloadArgParser(DownloadArgParser):
    def __init__(self,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)

        self.add_argument("-i", "--identifier",
                          help="Name of the output dataset where it's stored, overriding default",
                          default="mars",
                          type=str)
        self.add_argument("--compress",
                          help="Provide an integer from 1-9 (low to high) on how much to compress the output netCDF",
                          default=None,
                          type=int)

    def add_var_specs(self):
        super().add_var_specs()
        self.add_argument("-s", "--system",
                          help="ECMWF System identifier",
                          type=int,
                          default=5)
        self.add_argument("product_type",
                          help="ECMWF product type",
                          type=str,
                          default=None)
        self.add_argument("stream",
                          help="ECMWF product stream",
                          type=str,
                          default=None)
        self.add_argument("params",
                          help="This provides the necessary ECMWF parameter IDs for requests",
                          default=None,
                          type=csv_arg)
        self.add_argument("attributes",
                          help="This provides the necessary ECMWF attribute names to select data with",
                          default=None,
                          type=csv_arg)
        return self


class MARSDataset(DatasetConfig):
    def __init__(self,
                 *args,
                 identifier: str = None,
                 params: list = None,
                 attributes: list = None,
                 **kwargs):
        super().__init__(*args,
                         identifier="mars"
                         if identifier is None else identifier,
                         **kwargs)

        # Handle reverse mapping of keyword parameters back to arguments: it is a hack
        # but allows the reconstruction of the object
        if "mars_mapping" in kwargs and params is None and attributes is None:
            params = [v["param"] for v in kwargs["mars_mapping"].values()]
            attributes = [v["attribute"] for v in kwargs["mars_mapping"].values()]

        if not(len(params) == len(attributes) == len(self.var_prefixes)):
            raise AttributeError("The number of parameter strings, attributes and variables must match: {} vs {} vs {}"
                                 .format(len(params), len(attributes), len(self.variables)))

        self._mars_mapping = {var_prefix: dict(param=params[order], attribute=attributes[order])
                              for order, var_prefix in enumerate(self.var_prefixes)}

    @property
    def mars_mapping(self):
        return self._mars_mapping


class MARSDownloader(Downloader):
    def __init__(self,
                 dataset: MARSDataset,
                 product_type: str,
                 stream: str,
                 *args,
                 product_class: str = "od",
                 system: int = 5,
                 compress: Union[int, None] = None,
                 request_args: [dict, None] = None,
                 **kwargs):
        self._product_class = product_class
        self._product_type = product_type
        self._stream = stream
        self._system = system
        self._compress = compress
        self._request_args = request_args

        self._server = ecmwfapi.ECMWFService("mars")

        super().__init__(dataset,
                         *args,
                         # TODO: not true, this could be max HOUR, need to implement
                         source_min_frequency=Frequency.YEAR,
                         source_max_frequency=Frequency.MONTH,
                         **kwargs)

    def _single_download(self,
                         var_config: object,
                         req_dates: object) -> list:
        logging.debug("Processing {} dates for {}".format(len(req_dates), var_config))
        temp_download_path = os.path.join(var_config.root_path,
                                          self.dataset.location.name,
                                          "temp.{}".format(os.path.basename(
                                              self.dataset.var_filepath(var_config, req_dates))))
        download_path = os.path.join(var_config.root_path,
                                     self.dataset.location.name,
                                     os.path.basename(self.dataset.var_filepath(var_config, req_dates)))
        os.makedirs(os.path.dirname(download_path), exist_ok=True)

        if os.path.exists(download_path):
            logging.warning(f"We have a downloaded file available, skipping: {download_path}")
            return [download_path]

        # TODO: handle time
        request_template = """        retrieve,
          class={product_class},
          date={date},
          {request_args}expver=1,
          levtype={levtype},
          {levlist}param={params},
          method=1,
          origin=ecmf,
          stream={stream},
          system={system},
          time=00:00:00,
          type={product_type},
          area={area},
          grid=0.25/0.25,
          target="{target}",
          format=netcdf
            """
        #          step={step},

        levtype = "plev" if var_config.level else "sfc"

        if self.request_frequency <= Frequency.MONTH:
            req_dates = [dt.replace(day=1) for dt in req_dates]

        request = request_template.format(
            area="/".join([str(s) for s in self.dataset.location.bounds]),
            date="/".join([el.strftime("%Y%m%d") for el in req_dates]),
            levtype=levtype,
            levlist="levelist={},\n          ".format(var_config.level)
            if var_config.level else "",
            params=self.dataset.mars_mapping[var_config.prefix]['param'],
            product_class=self._product_class,
            product_type=self._product_type,
            request_args="".join(["{}={},\n          ".format(k, v) for k, v in self._request_args.items()])
                         if self._request_args is not None else "",
            # TODO: specify - limitations based on access rights to ecmwf data
            step=0,
            stream=self._stream,
            system=self._system,
            target=os.path.basename(temp_download_path),
        )

        if not os.path.exists(temp_download_path):
            logging.debug("MARS REQUEST: \n{}\n".format(request))
            #import sys; sys.exit(0)
            try:
                self._server.execute(request, temp_download_path)
            except ecmwfapi.api.APIException:
                logging.exception("Could not complete ECMWF request: {}")
                return []
        else:
            logging.debug("Already have {}".format(temp_download_path))

        logging.debug("Files downloaded: {}".format(temp_download_path))

        ds = xr.open_dataset(temp_download_path)
        da = getattr(ds, self.dataset.mars_mapping[var_config.prefix]['attribute'])

        if var_config.level:
            da = da.sel(level=int(var_config.level))

        logging.info("Saving MARS file to {}".format(download_path))
        xr_save_netcdf(da, download_path, complevel=self._compress)

        ds.close()

        #if os.path.exists(temp_download_path):
        #    logging.info("Removing {}".format(temp_download_path))
        #    os.unlink(temp_download_path)

        return [download_path]


def mars_main():
    args, request_args = (MARSDownloadArgParser().
                          add_var_specs().
                          add_workers().
                          parse_known_args())
    logging.info("MARS API data downloading")

    location = Location(
        name=args.hemisphere,
        north=args.hemisphere == "north",
        south=args.hemisphere == "south",
    )

    dataset = MARSDataset(
        levels=args.levels,
        location=location,
        var_names=args.vars,
        frequency=getattr(Frequency, args.frequency),
        output_group_by=getattr(Frequency, args.output_group_by),
        config_path=args.config,
        overwrite=args.overwrite_config,
        params=args.params,
        attributes=args.attributes,
    )

    for start_date, end_date in zip(args.start_dates, args.end_dates):
        logging.info("Downloading between {} and {}".format(start_date, end_date))
        mars = MARSDownloader(
            dataset,
            args.product_type,
            args.stream,
            start_date=start_date,
            end_date=end_date,
            max_threads=args.workers,
            request_frequency=getattr(Frequency, args.output_group_by),
            request_args=request_args,
            compress=args.compress,
        )
        mars.download()

        dataset.save_data_for_config(
            rename_var_list={a: v for a, v in zip(args.attributes, args.vars)},
            source_files=mars.files_downloaded,
            var_filter_list=["lambert_azimuthal_equal_area"],
        )



from abc import abstractmethod, ABCMeta
import logging
import os
import shutil

from download_toolbox.config import Configuration


class DataCollection(metaclass=ABCMeta):
    """An Abstract base class with common interface for data collection classes.

    It represents a collection of data assets on a filesystem, though in the future
    it would make sense that we also allow use for object storage etc.

    This also handles automatic egress/ingress and validation of the configurations
    for these collections.

    :param _identifier: The identifier of the data collection.
    :param _path: The base path of the data collection.
    :raises AssertionError: Raised if identifier is not specified, or no hemispheres are selected.
    """

    @abstractmethod
    def __init__(self,
                 *,
                 identifier: str,
                 base_path: str = os.path.abspath(os.path.join(".", "data")),
                 config_path: os.PathLike = None,
                 config_type: str = "data_collection",
                 path_components: list = None,
                 dummy: bool = False,
                 **kwargs) -> None:
        self._identifier = identifier

        path_components = list() if path_components is None else path_components
        if not isinstance(path_components, list):
            raise DataCollectionError("path_components should be an Iterator")

        # TODO: seriously: root_path, path and base_path!? Rationalise this, too smelly for words
        self._base_path = base_path
        self._dummy = dummy
        self._path_components = path_components
        self._root_path = None
        self._path = None
        self._config = None
        self._config_path = config_path
        self._config_type = config_type

        self.init()

    def copy_to(self,
                new_identifier: object,
                base_path: os.PathLike = None,
                skip_copy: bool = False):
        """

        Args:
            new_identifier:
            base_path:
            skip_copy:
        """
        old_path = self.path

        if base_path is not None:
            logging.info("Setting base path for copy to {}".format(base_path))
            self._base_path = base_path

        self.identifier = new_identifier

        if not skip_copy:
            logging.info("Copying {} to {}".format(old_path, self.path))
            shutil.copytree(old_path, self.path, dirs_exist_ok=True)
        else:
            logging.warning("Skipping naive copy of {} to {}".format(old_path, self.path))

    def get_config(self,
                   config_funcs: dict = None,
                   strip_keys: list = None) -> dict:
        """get_config returns the implementation configuration for re-instantiation

        get_config returns a configuration dictionary that provides not just a reference
        but also a portability layer for recreating classes.

        For things that aren't serialisable natively, use config_funcs to serialise or represent
        values that allow recreation (it's on you to recreate those appropriately). An example
        is available at ...download_toolbox.interface.get_dataset_config_implementation

        If you supply any arguments in a derived implementation, use strip_keys to prevent
        them being exported into configurations that would then result in duplicate arguments
        when the class is recreated from config

        TODO: documenting get_config in derived implementations
        TODO: schema and validation for this library and others, helping to control implementations
         to aid portability of pipelines

        Args:
            config_funcs:
            strip_keys:

        Returns:

        """
        strip_keys = [] if strip_keys is None else strip_keys
        return {k: config_funcs[k](v) if config_funcs is not None and k in config_funcs else v
                for k, v in self.__dict__.items() if k not in ["_path", "_config", "_config_path", "_root_path"] + strip_keys}

    def init(self):
        self._config = None
        self._root_path = os.path.join(self._base_path, self._identifier)
        self._path = os.path.join(self._root_path, *self._path_components)

        if self._identifier is None:
            raise DataCollectionError("No identifier supplied")

        if not self._dummy:
            if os.path.exists(self._path):
                logging.debug("{} already exists".format(self._path))
            else:
                if not os.path.islink(self._path):
                    logging.info("Creating path: {}".format(self._path))
                    os.makedirs(self._path, exist_ok=True)
                else:
                    logging.info("Skipping creation for symlink: {}".format(self._path))
        else:
            logging.debug("Avoiding creation of path {} as collection is marked dummy".format(self._path))

    def save_config(self):
        saved_config = self.config.render(self)
        logging.info("Saved dataset config {}".format(saved_config))
        return saved_config

    @property
    def base_path(self):
        return self._base_path

    @property
    def config(self):
        if self._config is None:
            self._config = Configuration(
                config_path=self._config_path if self._config_path is not None else self._root_path,
                config_type=self._config_type,
                identifier=self.identifier)
        return self._config

    @property
    def config_path(self):
        return self.config.output_path

    @config_path.setter
    def config_path(self, config_path: os.PathLike) -> None:
        self._config_path = config_path
        self._config.output_path = config_path

    @property
    def config_type(self):
        return self._config_type

    @property
    def identifier(self) -> str:
        """The identifier (label) for this data collection."""
        return self._identifier

    @identifier.setter
    def identifier(self, identifier: str) -> None:
        self._identifier = identifier
        self.init()

    @property
    def path(self) -> str:
        """The base path of the data collection."""
        return self._path

    @path.setter
    def path(self, path: str) -> None:
        self._path = path

    @property
    def path_components(self):
        return self._path_components

    @property
    def root_path(self):
        return self._root_path


#    def __repr__(self):
#        return "{} with path {}".format(self.name, self.path)


class DataCollectionError(RuntimeError):
    pass





import datetime as dt
import logging
import os
import sys

from collections import UserDict

import orjson


class ConfigurationError(RuntimeError):
    pass


class Configuration(UserDict):
    def __init__(self,
                 identifier,
                 config_type: str = "config",
                 config_path: os.PathLike = None,
                 **kwargs):
        super().__init__(kwargs)

        self._identifier = identifier
        self._history = []
        self._path = config_path
        self._config_type = config_type

        self._load_existing()

    def _load_existing(self):
        if os.path.exists(self.output_path):
            logging.info("Loading configuration {}".format(self.output_path))
            with open(self.output_path, "r") as fh:
                data = fh.read()
            obj = orjson.loads(data)
            self._history.extend(obj["history"])
            self.data.update(obj["data"])

    def render(self,
               owner,
               implementation=None):
        if self._path is not None:
            output_path = os.path.abspath(self._path)
            if not os.path.isdir(os.path.dirname(output_path)):
                raise ConfigurationError("Path {} should be path we can output to".format(self._path))

        self._history.append(" ".join([
            "Run at {}: ".format(dt.datetime.now(dt.timezone.utc).strftime("%c %Z")),
            *sys.argv]))

        configuration = {
            "data": owner.get_config(),
            "history": self._history,
            "implementation": implementation
            if implementation is not None
            else ":".join([owner.__module__, owner.__class__.__name__]),
        }

        logging.info("Writing configuration to {}".format(self.output_path))
        str_data = orjson.dumps(configuration, option=orjson.OPT_INDENT_2)

        with open(self.output_path, "w") as fh:
            fh.write(str_data.decode())
        return self.output_path

    @property
    def identifier(self):
        return self._identifier

    @property
    def output_path(self):
        default_filename = "{}.{}.json".format(self._config_type, self.identifier)
        return os.path.join(".", default_filename) if self._path is None else \
            os.path.join(self._path, default_filename) if os.path.isdir(self._path) else \
            self._path

    @output_path.setter
    def output_path(self, path: os.PathLike) -> None:
        self._path = path
